"""Using `Elegant`_ files with PyAT
==================================

PyAT can read lattice descriptions in Elegant format (.lte files), and can export
lattices in Elegant format.

However, because of intrinsic differences between PyAT and Elegant, some
incompatibilities must be taken into account.

1. Translation losses
---------------------

Multipoles
^^^^^^^^^^^^^^

Elegant thick multipoles are limited to a single multipole order. For AT
:py:class:`.Multipole` elements combining several orders, the lowest order is converted
and higher orders are discarded. AT :py:class:`.ThinMultipole` elements are expanded
to a series of thin ``MULT`` elements with 0 length.

Elegant elements absent from AT
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Many Elegant elements have no equivalent in AT. They are replaced by
:py:class:`.Marker` or :py:class:`.Drift` elements, depending on their length.

Incompatible attributes
^^^^^^^^^^^^^^^^^^^^^^^^^^^
Some AT element attributes have no Elegant equivalent, and vice versa.

When exporting to an Elegant file:

- `FringeBendEntrance`, `FringeBendExit`, `FringeQuadEntrance`, `FringeQuadExit` are
  discarded,
- `R1`, `R2`, `T1`, `T2` are discarded,
- `RApertures`, `EApertures` are discarded.

When reading an Elegant file:

- `TILT` is interpreted and converted to `R1` and `R2` attributes,
- `DX`, `DZ` are converted to `T1`and `T2` attributes,
- `N_SLICES`, if specified, is converted to `NumIntSteps`. If not, `NumIntSteps` is
  left to is default value (10).

2. Usage
--------

Usage Elegant files is similar to :ref:`using MAD-X files <using-mad-x-files>`. The
Elegant RPN calculator is emulated.

3. Functions and classes
------------------------

.. _elegant: https://ops.aps.anl.gov/manuals/elegant_latest/elegant.html#elegantse9.html
"""

from __future__ import annotations

__all__ = ["ElegantParser", "_ElegantExporter", "load_elegant", "save_elegant"]

import functools
from math import sqrt, factorial
from collections.abc import Iterable
import warnings

import numpy as np
from scipy.constants import c as clight

from .allfiles import register_format
from .file_input import ElementDescr, BaseParser, UpperCaseParser
from .file_input import skip_names, ignore_names, ignore_class
from .file_output import Exporter
from ..lattice import Particle, Lattice, elements as elt, tilt_elem, shift_elem

# noinspection PyProtectedMember
from .madx import sinc, _Line, p_dict, p_list
from . import Rpn


# -------------------
#  Utility functions
# -------------------


def elegant_element(func):
    """Decorator which tilts and shifts the decorated AT element"""

    @functools.wraps(func)
    def wrapper(
        self,
        *args,
        tilt=0.0,
        dx=0.0,
        dy=0.0,
        n_slices=None,
        n_kicks=None,
        **kwargs,
    ):
        if n_kicks is not None:  # Deprecated parameter
            kwargs["NumIntSteps"] = int(n_kicks / 4)
        if n_slices is not None:
            kwargs["NumIntSteps"] = n_slices
        elems = func(self, *args, **kwargs)
        for el in elems:
            if tilt != 0.0:
                tilt_elem(el, tilt)
            if not (dx == 0.0 and dy == 0.0):
                shift_elem(el, deltax=dx, deltaz=dy)
            el.origin = self.origin
        return elems

    return wrapper


# ------------------------------
#  Base class for Elegant elements
# ------------------------------


class _ElegantElement(ElementDescr):
    """Description of MADX elements"""

    str_attr = {"filename", "group", "mode", "insert_from"}


# ------------------------------
#  Elegant classes
# ------------------------------


# noinspection PyPep8Naming
class DRIF(_ElegantElement):
    @elegant_element
    def to_at(self, l=0.0, **params):  # noqa: E741
        return [elt.Drift(self.name, l, **params)]


# noinspection PyPep8Naming
class MARK(_ElegantElement):
    at2mad = {}

    @elegant_element
    def to_at(self, **params):
        return [elt.Marker(self.name, **params)]


# noinspection PyPep8Naming
class KQUAD(_ElegantElement):
    @elegant_element
    def to_at(self, l, k1=0.0, **params):  # noqa: E741
        return [elt.Quadrupole(self.name, l, k1, **params)]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["", "K1"], kwargs.pop("PolynomB", ())))
        return el


# noinspection PyPep8Naming
class KSEXT(_ElegantElement):
    @elegant_element
    def to_at(self, l, k2=0.0, **params):  # noqa: E741
        return [elt.Sextupole(self.name, l, k2 / 2.0, **params)]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["", "K1", "K2"], kwargs.pop("PolynomB", ())))
        return el


# noinspection PyPep8Naming
class KQUSE(_ElegantElement):
    @elegant_element
    def to_at(self, l, k1=0.0, k2=0.0, **params):  # noqa: E741
        poly_b = [0.0, k1, k2 / 2.0]
        return [elt.Multipole(self.name, l, [], poly_b, **params)]


# noinspection PyPep8Naming
class KOCT(_ElegantElement):
    @elegant_element
    def to_at(self, l, k3=0.0, **params):  # noqa: E741
        poly_b = [0.0, 0.0, 0.0, k3 / 6.0]
        poly_a = [0.0, 0.0, 0.0, 0.0]
        return [elt.Octupole(self.name, l, poly_a, poly_b, **params)]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["", "", "", "K3"], kwargs.pop("PolynomB", ())))
        return el


class MULT(_ElegantElement):
    at2mad = {"Length": "L", "Order": "ORDER", "Value": "KNL"}

    @elegant_element
    def to_at(self, l=0, knl=0.0, order=1, **params):  # noqa: E741
        poly_a = np.zeros(order + 1)
        poly_b = np.zeros(order + 1)
        if l == 0.0:
            poly_b[order] = knl / factorial(order)
            return [elt.ThinMultipole(self.name, poly_a, poly_b, **params)]
        else:
            poly_b[order] = knl / factorial(order) / l
            return [elt.Multipole(self.name, l, poly_a, poly_b, **params)]


# noinspection PyPep8Naming
class CSBEND(_ElegantElement):
    at2mad = {
        "Length": "L",
        "BendingAngle": "ANGLE",
        "EntranceAngle": "E1",
        "ExitAngle": "E2",
    }

    @elegant_element
    def to_at(
        self,
        l,  # noqa: E741
        angle,
        e1=0.0,
        e2=0.0,
        k1=0.0,
        k2=0.0,
        k3=0.0,
        k4=0.0,
        hgap=None,
        fint=0.0,
        **params,
    ):
        if hgap is not None:
            params.update(FullGap=2.0 * hgap, FringeInt1=fint, FringeInt2=fint)
        if k2 != 0.0 or k3 != 0.0 or k4 != 0.0:
            params["PolynomB"] = [0.0, k1, k2 / 2.0, k3 / 6.0, k4 / 24.0]
        return [
            elt.Dipole(
                self.name,
                l,
                angle,
                k1,
                EntranceAngle=e1,
                ExitAngle=e2,
                **params,
            )
        ]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        el.update(p_dict(["K0", "K1", "K2", "K3"], kwargs.pop("PolynomB", ())))
        return el


# noinspection PyPep8Naming
class RBEN(CSBEND):
    @elegant_element
    def to_at(self, l, angle, e1=0.0, e2=0.0, **params):  # noqa: E741
        hangle = abs(0.5 * angle)
        arclength = l / sinc(hangle)
        return super().to_at(arclength, angle, e1=hangle + e1, e2=hangle + e2, **params)

    @property
    def length(self):
        """Element length"""
        hangle = 0.5 * self["angle"]
        return self["l"] / sinc(hangle)


# noinspection PyPep8Naming
class KICKER(_ElegantElement):
    @elegant_element
    def to_at(self, l=0.0, hkick=0.0, vkick=0.0, **params):  # noqa: E741
        kicks = np.array([hkick, vkick], dtype=float)
        return [elt.Corrector(self.name, l, kicks, **params)]

    @classmethod
    def from_at(cls, kwargs):
        el = super().from_at(kwargs)
        kicks = kwargs.pop("KickAngle", (0.0, 0.0))
        el["HKICK"] = kicks[0]
        el["VKICK"] = kicks[1]
        return el


# noinspection PyPep8Naming
class HKICK(KICKER):
    @elegant_element
    def to_at(self, l=0.0, kick=0.0, **params):  # noqa: E741
        return super().to_at(l=l, hkick=kick, **params)


# noinspection PyPep8Naming
class VKICK(KICKER):
    @elegant_element
    def to_at(self, l=0.0, kick=0.0, **params):  # noqa: E741
        return super().to_at(l=l, vkick=kick, **params)


# noinspection PyPep8Naming
class RFCA(_ElegantElement):
    at2mad = {"Length": "L", "Voltage": "VOLT", "Frequency": "FREQ"}

    @elegant_element
    def to_at(
        self,
        l=0.0,  # noqa: E741
        volt=0.0,
        freq=np.nan,
        **params,
    ):
        cavity = elt.RFCavity(
            self.name,
            l,
            volt,
            freq,
            0,
            0.0,
            PassMethod="IdentityPass" if l == 0.0 else "DriftPass",
            **params,
        )
        return [cavity]


# noinspection PyPep8Naming
class MONI(_ElegantElement):
    @elegant_element
    def to_at(self, l=0.0, **params):  # noqa: E741
        if l == 0.0:
            return [elt.Monitor(self.name, **params)]
        else:
            hl = 0.5 * l
            return [
                elt.Drift(self.name, hl, origin="MONI"),
                elt.Monitor(self.name, **params),
                elt.Drift(self.name, hl, origin="MONI"),
            ]


# noinspection PyPep8Naming
class HMON(MONI):
    pass


# noinspection PyPep8Naming
class VMON(MONI):
    pass


def multipole(kwargs):
    """AT ThinMultipole or Multipole converted to Elegant MULT"""

    def singlemul(nm, o, v):
        if nm is None:
            nm = ".".join((name, str(o)))
        return MULT.from_at({"FamName": nm, "Length": length, "Order": o, "Value": v})

    name = kwargs.pop("FamName")
    length = kwargs.pop("Length", 0.0)
    poly_b = p_list(kwargs.pop("PolynomB", ()))
    if length == 0.0:
        return [singlemul(None, o, v) for o, v in enumerate(poly_b) if v != 0.0]
    else:
        for order, v in enumerate(poly_b):
            if v != 0.0:
                return singlemul(name, order, length * v)
        return singlemul(name, len(poly_b) - 1, 0.0)


def ignore(kwargs):
    """AT element ignored in Elegant: convert to marker or drift"""
    length = kwargs.get("Length", 0.0)
    if length == 0.0:
        print(f"{kwargs['name']} is replaced by a marker")
        return MARK.from_at(kwargs)
    else:
        print(f"{kwargs['name']} is replaced by a drift")
        return DRIF.from_at(kwargs)


SOLENOID = ignore_class("SOLENOID", _ElegantElement)

skip_names(
    globals(),
    _ElegantElement,
    [
        "MAXAMP",
        "CHARGE",
        "RECIRC",
        "MALIGN",
        "SREFFECTS",
        "PFILTER",
        "ENERGY",
        "SCATTER",
        "WATCH",
        "WAKE",
    ],
)


ignore_names(globals(), _ElegantElement, ["SCRAPER", "ECOL", "RCOL", "CSRDRIF"])

EDRIFT = DRIFT = DRIF
QUAD = QUADRUPOLE = KQUAD
CSRCSBEN = CSBEN = SBEN = SBEND = CSBEND
CRBEN = CRBEND = RBEND = RBEN
SEXT = SEXTUPOLE = KSEXT
OCTU = OCTUPOLE = KOCT
HKICKER = HKICK
VKICKER = VKICK
MONITOR = MONI
HMONITOR = HMON
VMONITOR = VMON
RFCW = RFCA
SOLE = SOLENOID
MULTIPOLE = MULT


class ElegantParser(UpperCaseParser, BaseParser):
    # noinspection PyUnresolvedReferences
    """Elegant parser

    The parser is a subclass of :py:class:`dict` and is a database containing all the
    Elegant objects.

    Example:
        Parse a file:

        >>> parser = at.ElegantParser()
        >>> parser.parse_file("file1")

        Look at the "qf1" element

        >>> parser["QF1"]
        QUAD(name=QF1, l=1.0, k1=0.5, tilt=0.001)

        Generate an AT :py:class:`.Lattice` from the "RING" sequence

        >>> ring = parser.lattice(use="RING")  # generate an AT Lattice
    """

    _continuation = "&"
    _linecomment = "!"
    _endfile = "RETURN"

    def __init__(self, **kwargs):
        """
        Args:
            verbose:    If :py:obj:`True`, print details on the processing
            **kwargs:   Initial variable definitions
        """
        super().__init__(globals(), **kwargs)
        self.rpn = Rpn()

    def _assign(self, label: str, key: str, val: str):
        # Special treatment of "line=(...)" commands
        if key == "LINE":
            val = val.replace(")", ",)")  # For tuples with a single item
            return label, _Line(self.evaluate(val), name=label)
        else:
            return super()._assign(label, key, val)

    def _command(self, label, cmdname, *args: str, **kwargs):
        # Special treatment of label #INCLUDE
        if label == "#INCLUDE":
            file = cmdname[1:-1] if cmdname[0] == '"' else cmdname
            self.parse_files(file, final=False)
        else:
            return super()._command(label, cmdname, *args, **kwargs)

    def _statement(self, line: str) -> bool:
        # Special treatment of RPN definitions
        if line[0] == "%":
            self.rpn.input(line[1:])
            return True
        else:
            return super()._statement(line)

    def _argparser(
        self,
        argcount: int,
        argstr: str,
        *,
        bool_attr: tuple[str] = (),
        str_attr: tuple[str] = (),
        pos_args: tuple[str] = (),
    ):
        """Evaluate a command argument and return a pair (key, value)"""

        def arg_value(k, v):
            if k in str_attr:
                return v[1:-1] if v[0] == '"' else v
            elif v[0] == '"':
                return self.rpn.evaluate(v[1:-1])
            else:
                return self.evaluate(v)

        key, *value = argstr.split(sep="=", maxsplit=1)
        if value:  # Keyword argument
            key = key.lower()
            return key, arg_value(key, value[0])
        else:
            try:
                key = pos_args[argcount]
            except IndexError:
                print(f"Unexpected positional argument '{argstr}' ignored")
                return None
            return key, arg_value(key, argstr)

    def _generator(self, params: dict) -> Iterable[elt.Element]:
        def beta() -> float:
            rest_energy = params["particle"].rest_energy
            if rest_energy == 0.0:
                return 1.0
            else:
                gamma = float(params["energy"] / rest_energy)
                return sqrt(1.0 - 1.0 / gamma / gamma)

        cavities = []
        cell_length = 0

        for elem in super()._generator(params):
            if isinstance(elem, elt.RFCavity):
                cavities.append(elem)
            cell_length += getattr(elem, "Length", 0.0)
            yield elem

        part = params.setdefault("particle", "relativistic")
        if isinstance(part, str):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                params["particle"] = Particle(part)

        rev = beta() * clight / cell_length

        # Set the cavities' Energy and harmonic number
        for cav in cavities:
            h = cav.Frequency / rev
            cav.HarmNumber = h


def load_elegant(
    *files: str, use: str = "RING", verbose: bool = False, **kwargs
) -> Lattice:
    """Create a :py:class:`.Lattice` from Elegant lattice files

    - Elegant lattice files do not specify the beam energy. :py:class:`ElegantParser`
      sets it by default to 1.0 GeV. Use the *energy* keyword to set it to the
      desired value.
    - Long elements are split according to the default AT value of *NumIntSteps* (10)
      unless *N_SLICES* is specified in the Elegant element definition.

    Parameters:
        files:              Names of one or several Elegant lattice description files
        use:                Name of the Elegant LINE describing the desired
          lattice. Default: ``RING``
        verbose:            If :py:obj:`True`, print details on the processing

    Keyword Args:
        name (str):         Name of the lattice. Default: Elegant sequence name
        particle(Particle): Circulating particle. Default: Particle("relativistic")
        energy (float):     Energy of the lattice [eV]. Default: 1.0E9
        periodicity(int):   Number of periods. Default: 1
        *:                  Other keywords will be used as Lattice attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    parser = ElegantParser(verbose=verbose)
    parser.parse_files(*files)
    return parser.lattice(use=use, **kwargs)


_AT2EL = {
    elt.Quadrupole: KQUAD.from_at,
    elt.Sextupole: KSEXT.from_at,
    elt.Octupole: KOCT.from_at,
    elt.ThinMultipole: multipole,
    elt.Multipole: multipole,
    elt.RFCavity: RFCA.from_at,
    elt.Drift: DRIF.from_at,
    elt.Bend: SBEN.from_at,
    elt.Marker: MARK.from_at,
    elt.Monitor: MONI.from_at,
    elt.Corrector: KICKER.from_at,
}


class _ElegantExporter(Exporter):
    delimiter = ""
    continuation = "&"
    bool_fmt = {False: ".FALSE.", True: ".TRUE."}
    use_line = True

    def generate_madelems(
        self, eltype: type[elt.Element], elemdict: dict
    ) -> ElementDescr | list[ElementDescr]:
        return _AT2EL.get(eltype, ignore)(elemdict)


def at2elegant(attype):
    return _AT2EL.get(attype, ignore)


def save_elegant(
    ring: Lattice,
    filename: str | None = None,
    **kwargs,
):
    """Save a :py:class:`.Lattice` as an Elegant file

    Args:
        ring:   lattice
        filename: file to be created. If None, write to sys.stdout

    Keyword Args:
        use (str | None): name of the created SEQUENCE of LINE.
          Default: name of the PyAT lattice
    """
    exporter = _ElegantExporter(ring, **kwargs)
    exporter.export(filename)


register_format(
    ".lte",
    load_elegant,
    save_elegant,
    descr="Elegant lattice description. See :py:func:`.load_elegant`.",
)



"""Global set of constants."""

from __future__ import annotations

__all__ = ["test_mode", "set_test_mode", "DConstant", "random"]

import multiprocessing
import os
from contextlib import contextmanager
from contextvars import ContextVar

from numpy.random import Generator, PCG64, SeedSequence

from ..cconfig import ismpi, isopenmp, iscuda, isopencl

if ismpi():
    from mpi4py import MPI

    _comm = MPI.COMM_WORLD
    _MPI_sz = _comm.Get_size()
    _MPI_rk = _comm.Get_rank()
else:
    _MPI_sz = 1
    _MPI_rk = 0

_test_mode: ContextVar[int] = ContextVar("test_mode", default=0)


@contextmanager
def set_test_mode(mode: int = 1) -> None:
    r"""Context manager to set a \"test mode\" integer value.

    Args:
        mode:   a non-zero value which will be returned by the :py:func:`test_mode`
          function.

    Example:
        >>> assert not test_mode(), "test mode should not be set"
        >>> with set_test_mode():
        ...     assert test_mode(), "test mode should be set"

        Test mode contexts can be nested:

        >>> with set_test_mode(1):
        ...     assert test_mode() == 1, "outer context"
        ...     with set_test_mode(2):
        ...         assert test_mode() == 2, "inner context"
        ...     assert test_mode() == 1, "back to outer context"
    """
    tok = _test_mode.set(mode)
    try:
        yield None
    finally:
        _test_mode.reset(tok)


def test_mode() -> int:
    """Return the test mode value.

    Returns:
        test_mode:  0 outside a :py:func:`set_test_mode` context. Within such context,
          return its non-zero value
    """
    return _test_mode.get()


def _newgen(seed=12345):
    ss = SeedSequence(seed)
    seeds = ss.spawn(_MPI_sz + 1)
    return Generator(PCG64(seeds[0])), Generator(PCG64(seeds[_MPI_rk + 1]))


class _Dst:
    """Set of constants for AT numerical analysis."""

    # optimized values:
    # see https://github.com/atcollab/at/pull/199#issuecomment-745420443
    XYStep: float = 3.0e-8  # Coordinate step for differentiation
    DPStep: float = 3.0e-6  # Momentum step for dispersion and chromaticity
    OrbConvergence: float = 1.0e-12  # Convergence criterion for orbit
    OrbMaxIter: int = 20  # Max. number of iterations for orbit
    TStol: float = 1.0e-9  # Tolerance for synchronous phase search
    omp_num_threads: int = int(os.environ.get("OMP_NUM_THREADS", "0"))
    patpass_poolsize: int = multiprocessing.cpu_count()
    patpass_startmethod: str | None = None
    _rank: int = _MPI_rk  # MPI rank

    def __setattr__(self, name: str, value):
        _ = getattr(self, name)  # make sure attribute exists
        object.__setattr__(self, name, value)

    def reset(self, name: str) -> None:
        delattr(self, name)

    @property
    def mpi(self) -> bool:
        return ismpi()

    @property
    def openmp(self) -> bool:
        return isopenmp()

    @property
    def cuda(self) -> bool:
        return iscuda()

    @property
    def opencl(self) -> bool:
        return isopencl()

    @property
    def rank(self) -> int:
        return self._rank


class _Random:
    """Random generators for AT."""

    def __init__(self):
        self.reset()

    @classmethod
    def reset(cls, seed=12345):
        cls.common, cls.thread = _newgen(seed)


DConstant = _Dst()
"""
Default values for AT algorithms

Attributes:
    XYStep:              Coordinate step for differentiation
    DPStep:              Momentum step for dispersion and chromaticity
    OrbConvergence:      Convergence criterion for orbit
    OrbMaxIter:          Max. number of iterations for orbit
    TStol:               Tolerance for synchronous phase search
    omp_num_threads:     Default number of OpenMP threads
    patpass_poolsize:    Default size of multiprocessing pool
    patpass_startmethod: Default start method for the multiprocessing
    mpi:                 :py:obj:`True` if MPI is active
    openmp:              :py:obj:`True` if OpenMP is active
    cuda:                :py:obj:`True` if CUDA is active
    opencl:              :py:obj:`True` if OpenCL is active

Methods:
    reset(attrname):    Reset the attribute to its default value
"""

random = _Random()
"""
Random generators for AT

Attributes:
    common(:py:class:`~numpy.random.Generator`): random generator common to
        all threads
    thread(:py:class:`~numpy.random.Generator`): random generator specific to
        each thread

Methods:
    reset(seed=None):   Reset the generators.

Examples:
    >>> at.random.common.random()
    0.8699988509120198

    Generates a random number in [0, 1) which is identical in all threads

    >>> at.random.thread.normal(size=5)
    array([-0.02326275, -0.89806482, -1.69086929, -0.74399398,  0.70156743])

    Generates an array of Gaussian random values which is specific to
    each thread
"""



import os
from tempfile import mktemp

import pytest
from numpy.testing import assert_allclose

from at.lattice import Lattice, elements as elt


@pytest.fixture
def simple_hmba(hmba_lattice):
    """Modify hmba_lattice to make it compatible with MAD-X and Elegant"""
    ring = hmba_lattice.deepcopy()
    # Set NumIntSteps to default, remove FringeQuadEntrance and FringeQuadExit
    for mult in ring.select(elt.Multipole):
        mult.NumIntSteps = 10
        if isinstance(mult, (elt.Dipole, elt.Quadrupole)):
            del mult.FringeQuadEntrance
            del mult.FringeQuadExit
    # Replace Multipoles by Octupoles
    for id in ring.get_uint32_index("OF*"):
        q = ring[id]
        ring[id] = elt.Octupole(q.FamName, q.Length, q.PolynomA, q.PolynomB)
    # Disable useless e;ements
    for id in ring.get_uint32_index("SH*"):
        q = ring[id]
        ring[id] = elt.Drift(q.FamName, q.Length)
    return ring


@pytest.mark.parametrize("lattice", ["dba_lattice", "simple_hmba"])
@pytest.mark.parametrize(
    "suffix, options",
    (
        (".m", {}),
        (".repr", {}),
        (".mat", {"use": "abcd"}),
        (".mat", {"mat_key": "efgh"}),
        (".json", {}),
        (".seq", {"use": "ring"}),
        (".lte", {"use": "ring"}),
    ),
)
def test_m(request, lattice, suffix, options):
    ring0 = request.getfixturevalue(lattice)
    fname = mktemp(suffix=suffix)

    # Create a new file
    ring0.save(fname, **options)

    # load the new file
    ring1 = Lattice.load(fname, **options)

    # Check that we get back the original lattice
    el1, rg1, _ = ring0.linopt6()
    el2, rg2, _ = ring1.linopt6()
    assert_allclose(rg1.tune, rg2.tune, atol=1.0e-12)
    assert_allclose(rg1.chromaticity, rg2.chromaticity, atol=1.0e-12)

    os.unlink(fname)



from math import sqrt  # noqa: F401

import numpy as np
import pytest
from numpy.testing import assert_allclose as assert_close

from at import UnorderedParser, MadxParser, Mad8Parser, ElegantParser, set_argparser

__all__ = []

test_data1 = """
/* test block comment
delimiter:      ";"
linecomment:    "!", "//"
end of comments */

v1 = 42.0 ;                  // Test line comment
v2 = 2*(sqrt(4) + 2**2) ;    // Test expression

label : command1, flag1,     ! Test continuation
-FLAG2, title="test parser", ! Test disabled flag
arg1=v1, ARG2=v2, ARG3=V3 ;  ! Test postponed definition

V3 = True ; V4 = False ;     ! Test several commands
"""

test_data2 = r"""
/* test block comment
delimiter:      end-of-line
linecomment:    "#"
continuation:   "\"
end of comments */

V1 = 42.0                     # Test line comment
V2 = 2*(sqrt(4) + 2**2)       # Test expression

label : command1, flag1, \    # Test continuation
-FLAG2, title="test parser",\ # Test disabled flag
arg1=V1, ARG2=V2, ARG3=V3     # Test postponed definition

V3 = True ; V4 = False        # Test several commands

TEST: TWISS, D=BETA;
"""

madx_data = """
/* test block comment
delimiter:      ";"
linecomment:    "!", "//"
end of comments */

BEAM,   PARTICLE="electron", ENERGY=6.0, SEQUENCE=RING.1;
BEAM,   PARTICLE="positron", ENERGY=2.0;
BEAM,   SEQUENCE=RING.1, RADIATE;

! Test comment

Q1:     QUADRUPOLE, L:=QL ;         ! check forward reference

Q1.F:   Q1,         K1=0.5 ;
Q1.D:   Q1,         K1:=-Q1.F->K1;  ! check attribute access
        Q1.F,       TILT=0.001;     ! check element update
SOL1:   SOLENOID,   L=0.5,
                    K1S=3.0;        ! check continuation
MULT1:  MULTIPOLE,  KNL={1.0, 1.0, 2.0, 6.0};
HK1:    HKICKER,    L=0,    KICK=0.001;
VK1:    VKICKER,    L=0,    KICK=-0.001;
HVK2:   KICKER      L=0     VKICK:=VKICK;  ! Check spaces as separators
BPM1:   MONITOR,    L=0.1;
BPM2:   VMONITOR;
RF:     RFCAVITY,   VOLT=5, FREQ=352.2, HARMON=31;

CELL.1: SEQUENCE,   L=4.0;
QFCELL: Q1.F,       AT=0.75;
        Q1.D,       AT=2.0, FROM=QFCELL;
ENDSEQUENCE;

RING.1: SEQUENCE,   L=LRING;
        CELL.1,     AT=2.0;
        SOL1,       AT=4.5;
        MULT1,      AT=4.8;
        HK1,        AT=4.8;
        VK1,        AT=4.8;
        HVK2,       AT=4.8;
        BPM1,       AT=4.9;
        BPM2,       AT=5;
        CELL.1,     AT=7.0;
ENDSEQUENCE;

VALUE, EMASS;

QL = sqrt(4)/2;          ! check arithmetic
const real lring = 9.0;
VKICK = 0.003;

TEST: TWISS, D=BETA;
"""

mad8_data = """
comment
test block comment
delimiter:      ";"
linecomment:    "!", "//"
endcomment

BEAM,   PARTICLE="positron", ENERGY=2.0;

! Test comment

Q1:     QUADRUPOLE, L:=QL ;         ! check forward reference

Q1.F:   Q1,         K1=0.5 ;
Q1.D:   Q1,         K1:=-Q1.F[K1];  ! check attribute access
        Q1.F,       TILT=0.001;     ! check element update
SOL1:   SOLENOID,   L=0.5,
                    KS=3.0;         ! check continuation
MULT1:  MULTIPOLE,  KNL={1.0, 1.0, 2.0, 6.0};
HK1:    HKICKER,    L=0,    KICK=0.001;
VK1:    VKICKER,    L=0,    KICK=-0.001;
HVK2:   KICKER,     L=0,    VKICK=VKICK;
BPM1:   MONITOR,    L=0.1;
BPM2:   VMONITOR;
RF:     RFCAVITY,   VOLT=5, FREQ=352.2, HARMON=31;

CELL.1: SEQUENCE,   L=4.0;
QFCELL: Q1.F,       AT=0.75;
        Q1.D,       AT=2.0, FROM=QFCELL;
ENDSEQUENCE;

RING.1: SEQUENCE,   L=LRING;
        CELL.1,     AT=2.0;
        SOL1,       AT=4.5;
        MULT1,      AT=4.8;
        HK1,        AT=4.8;
        VK1,        AT=4.8;
        HVK2,       AT=4.8;
        BPM1,       AT=4.9;
        BPM2,       AT=5;
        CELL.1,     AT=7.0;
ENDSEQUENCE;

VALUE, EMASS;

QL = sqrt(4)/2;          ! check arithmetic
LRING = 9.0;
VKICK = 0.003;

TEST: TWISS, D=BETA;
"""


elegant_data = """
! Test comment

DR1:    DRIFT,      L=0.25
DR2:    DRIFT,      L=1.0
DR3:    DRIFT,      L=0.75
DR4:    DRIFT,      L=0.05
Q1.F:   QUADRUPOLE, L="2 2 /", K1=0.5, TILT=0.001
Q1.D:   QUADRUPOLE, L=1.0, K1=-0.5 ! check attribute access
SOL1:   SOLENOID,   L=0.5, &
                    K1S=3.0        ! check continuation
MULT1:  MULTIPOLE,  KNL=6.0, ORDER=3
"HK:1": HKICKER,    L=0,    KICK=0.001
"VK:1": VKICKER,    L=0,    KICK=-0.001
HVK2:   KICKER,     L=0,    VKICK=0.003
BPM1:   MONITOR,    L=0.1
BPM2:   VMONITOR
RF:     RFCAVITY,   VOLT=5.0E6, FREQ=352.2E6, HARMON=31

CELL.1: LINE=(DR1, Q1.F, DR2, Q1.D, DR3)

RING.1: LINE=(CELL.1, DR1, SOL1, DR4, MULT1, &
              HK:1, VK:1, HVK2, DR4, BPM1, DR4, BPM2, CELL.1)
"""
# Global definitions for BaseParser

true = True
false = False


def _command1_parser(parser, argcount, argstr):
    return parser._argparser(
        argcount, argstr, bool_attr={"flag1", "FLAG2"}, str_attr={"title"}
    )


@set_argparser(_command1_parser)
def command1(**kwargs):
    """Sample command testing that arguments are as expected"""
    assert kwargs["title"] == "test parser"
    assert kwargs["flag1"] is True
    assert kwargs["FLAG2"] is False
    assert kwargs["arg1"] == 42.0
    assert kwargs["ARG2"] == 12.0
    assert kwargs["ARG3"] is True
    return "done"


# End of global definitions


@pytest.mark.parametrize(
    "cont, delimit, linecomm, data",
    [[None, ";", ("!", "//"), test_data1], ["\\", ";", "#", test_data2]],
)
def test_unordered_parser(cont, delimit, linecomm, data: str):
    class _Parser(UnorderedParser):
        _blockcomment = ("/*", "*/")
        _linecomment = linecomm
        _continuation = cont
        _delimiter = delimit

    parser = _Parser(globals())
    parser.parse_lines(data.splitlines())
    assert parser["label"] == "done"  # Make sure that "command1" was executed


@pytest.mark.parametrize(
    "Parser, data, energy, particle, radiate",
    [
        [MadxParser, madx_data, 6.0e9, "electron", True],
        [Mad8Parser, mad8_data, 2.0e9, "positron", False],
    ],
)
def test_madx_parser(Parser, data: str, energy: float, particle: str, radiate: bool):
    # fmt: off
    expected_ring_pos = np.array([
        0.0,    0.25,   1.25,   2.25,   3.25,   4.25,   4.75,   4.8,    4.8,
        4.8,    4.8,    4.8,    4.85,   4.9,    4.9,    4.95,   5.0,    5.0,
        5.25,   6.25,   7.25,   8.25,   9.0,
    ])
    expected_cell_pos = np.array([0.,   0.25,   1.25,   2.25,   3.25,   4.])
    # fmt: on
    parser = Parser()
    parser.parse_lines(data.splitlines())

    sequences = parser.sequences

    assert sequences == ["cell_1", "ring_1"]

    ring = parser.lattice(use="ring.1")
    assert len(ring) == 22
    assert ring.energy == energy
    assert str(ring.particle) == particle
    assert ring.circumference == 9.0
    assert ring.is_6d is radiate
    assert_close(ring.get_s_pos(), expected_ring_pos)

    cell = parser.lattice(use="cell.1")
    assert len(cell) == 5
    assert cell.energy == 2.0e09
    assert str(cell.particle) == "positron"
    assert cell.circumference == 4.0
    assert cell.is_6d is False
    assert_close(cell.get_s_pos(), expected_cell_pos)


def test_elegant_parser():
    # fmt: off
    expected_ring_pos = np.array([
        0.0, 0.25, 1.25, 2.25, 3.25, 4.0, 4.25, 4.75, 4.8, 4.8,
        4.8,  4.8,  4.8, 4.85,  4.9, 4.9, 4.95,  5.0, 5.0,
        5.25, 6.25, 7.25, 8.25, 9.0,
    ])
    expected_cell_pos = np.array([0., 0.25, 1.25, 2.25, 3.25, 4.])
    # fmt: on
    parser = ElegantParser()
    parser.parse_lines(elegant_data.splitlines())

    sequences = parser.sequences

    assert sequences == ["CELL_1", "RING_1"]

    ring = parser.lattice(use="ring.1", particle="electron", energy=6.0e9)
    assert len(ring) == 23
    assert ring.energy == 6.0e9
    assert str(ring.particle) == "electron"
    assert ring.circumference == 9.0
    assert ring.is_6d is False
    assert_close(ring.get_s_pos(), expected_ring_pos)

    cell = parser.lattice(use="cell.1", particle="positron", energy=2.0e9)
    assert len(cell) == 5
    assert cell.energy == 2.0e09
    assert str(cell.particle) == "positron"
    assert cell.circumference == 4.0
    assert cell.is_6d is False
    assert_close(cell.get_s_pos(), expected_cell_pos)



"""
Non-linear optics
"""

import numpy as np
from typing import Optional, Sequence
from scipy.special import factorial
from ..lattice import Element, Lattice
from ..tracking import internal_lpass
from .orbit import Orbit, find_orbit
from .linear import get_tune, get_chrom, linopt6
from .harmonic_analysis import get_tunes_harmonic

__all__ = ["detuning", "chromaticity", "gen_detuning_elem", "tunes_vs_amp"]


def tunes_vs_amp(
    ring: Lattice,
    amp: Optional[Sequence[float]] = None,
    dim: Optional[int] = 0,
    nturns: Optional[int] = 512,
    **kwargs,
) -> np.ndarray:
    r"""Generates a range of tunes for varying x, y, or z amplitudes

    Parameters:
        ring:               Lattice description
        amp:                Oscillation amplitude. If amp is None the
                            tunes are calculated from :py:func:`.linopt6`
                            else tracking is used
        dim:                Coordinate index of the oscillation
        nturns:             Number of turns
        method:             ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:      Number of harmonic components to compute (before
                            mask applied)
        fmin:               Lower bound for tune
        fmax:               Upper bound for tune
        hann:               Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        tunes (ndarray):    Array of tunes qx, qy with shape (len(amp), 2)
    """

    def _gen_part(ring, amp, dim, orbit, ld, nturns):
        invx = np.array(
            [
                [1 / np.sqrt(ld["beta"][0]), 0],
                [ld["alpha"][0] / np.sqrt(ld["beta"][0]), np.sqrt(ld["beta"][0])],
            ]
        )

        invy = np.array(
            [
                [1 / np.sqrt(ld["beta"][1]), 0],
                [ld["alpha"][1] / np.sqrt(ld["beta"][1]), np.sqrt(ld["beta"][1])],
            ]
        )
        part = np.array([orbit] * len(amp)).T + 1.0e-6
        part[dim, :] += amp
        part = internal_lpass(ring, part, nturns=nturns)
        sh = part.shape
        partx = np.reshape(part[0, :], (sh[1], sh[3]))
        partxp = np.reshape(part[1, :], (sh[1], sh[3]))
        party = np.reshape(part[2, :], (sh[1], sh[3]))
        partyp = np.reshape(part[3, :], (sh[1], sh[3]))
        px = np.array([np.matmul(invx, [partx[i], partxp[i]]) for i in range(len(amp))])
        py = np.array([np.matmul(invy, [party[i], partyp[i]]) for i in range(len(amp))])
        return px[:, 0, :] - 1j * px[:, 1, :], py[:, 0, :] - 1j * py[:, 1, :]

    l0, bd, _ = linopt6(ring)
    orbit = l0["closed_orbit"]
    tunes = bd["tune"]

    if amp is not None:
        partx, party = _gen_part(ring, amp, dim, orbit, l0, nturns)
        qx = get_tunes_harmonic(partx, **kwargs)
        qy = get_tunes_harmonic(party, **kwargs)
        tunes = np.vstack((qx, qy)).T

    return np.array(tunes)


def detuning(
    ring: Lattice,
    xm: Optional[float] = 0.3e-4,
    ym: Optional[float] = 0.3e-4,
    npoints: Optional[int] = 3,
    nturns: Optional[int] = 512,
    **kwargs,
):
    """Computes the tunes for a sequence of amplitudes

    This function uses :py:func:`tunes_vs_amp` to compute the tunes for
    the specified amplitudes. Then it fits this data and returns
    the detuning coefficiant dQx/Jx, dQy/Jx, dQx/Jy, dQy/Jy and the
    qx, qy arrays versus x, y arrays

    Parameters:
        ring:       Lattice description
        xm:         Maximum x amplitude
        ym:         Maximum y amplitude
        npoints:    Number of points in each plane
        nturns:     Number of turns for tracking
        method:     ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:  Number of harmonic components to compute
                       (before mask applied)
        fmin:       Lower bound for tune
        fmax:       Upper bound for tune
        hann:       Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        q0 (ndarray): qx, qy from horizontal and vertical amplitude scans.
                      Fit results with shape (2, 2). Only the fractional part
                      is returned
        q1 (ndarray): Amplitude detuning coefficients with shape (2, 2)
                      ``[[dQx/dJx, dQy/dJx], [dQx/dJy, dQy/dJy]]``
        x (ndarray): x amplitudes (npoints, )
        q_dx (ndarray): qx, qy tunes as a function of x amplitude (npoints, 2)
                        Only the fractional parts are returned
        y (ndarray): y amplitudes (npoints, )
        q_dy (ndarray): qx, qy tunes as a function of y amplitude (npoints, 2)
                        Only the fractional parts are returned
    """
    lindata0, _, _ = linopt6(ring)
    gamma = (1 + lindata0.alpha * lindata0.alpha) / lindata0.beta

    x = np.linspace(-xm, xm, npoints)
    y = np.linspace(-ym, ym, npoints)
    x2 = x * x
    y2 = y * y

    q_dx = tunes_vs_amp(ring, amp=x, dim=0, nturns=nturns, **kwargs)
    q_dy = tunes_vs_amp(ring, amp=y, dim=2, nturns=nturns, **kwargs)
    q_dx, _ = np.modf(q_dx * ring.periodicity)
    q_dy, _ = np.modf(q_dy * ring.periodicity)

    idx = np.isfinite(q_dx[:, 0]) & np.isfinite(q_dx[:, 1])
    idy = np.isfinite(q_dy[:, 0]) & np.isfinite(q_dy[:, 1])

    fx = np.polyfit(x2[idx], q_dx[idx], 1)
    fy = np.polyfit(y2[idy], q_dy[idy], 1)

    q0 = np.array([[fx[1, 0], fx[1, 1]], [fy[1, 0], fy[1, 1]]])
    q1 = np.array(
        [
            [2 * fx[0, 0] / gamma[0], 2 * fx[0, 1] / gamma[0]],
            [2 * fy[0, 0] / gamma[1], 2 * fy[0, 1] / gamma[1]],
        ]
    )

    return q0, q1, x, q_dx, y, q_dy


def chromaticity(
    ring: Lattice,
    method: Optional[str] = "linopt",
    dpm: Optional[float] = 0.02,
    npoints: Optional[int] = 11,
    order: Optional[int] = 3,
    dp: Optional[float] = 0,
    **kwargs,
):
    r"""Computes the non-linear chromaticities

    This function computes the tunes for the specified momentum offsets.
    Then it fits this data and returns the chromaticity up to the given
    order (npoints>order)

    Parameters:
        ring:       Lattice description
        method:     ``'linopt'`` (dfault) returns the tunes from the
          :py:func:`linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with fft,

          ``'laskar'`` tracks a single particle and computes the tunes with
          NAFF.
        dpm:        Maximum momentum deviation
        npoints:    Number of momentum deviations
        order:      Order of the fit
        dp:         Central momentum deviation

    Returns:
        fit (ndarray): Horizontal Vertical polynomial coefficients
                       from ``np.polyfit`` of shape (2, order+1)
        dpa: dp array of shape (npoints,)
        qz: tune array of shape (npoints, 2)
    """
    if order == 0:
        return get_chrom(ring, dp=dp, **kwargs)
    elif order > npoints - 1:
        raise ValueError("order should be smaller than npoints-1")
    else:
        dpa = np.linspace(-dpm, dpm, npoints)
        qz = []
        for dpi in dpa:
            qz.append(
                get_tune(ring, method=method, dp=dp + dpi, remove_dc=True, **kwargs)
            )
        fit = np.polyfit(dpa, qz, order)[::-1]
        fitx = fit[:, 0] * factorial(np.arange(order + 1))
        fity = fit[:, 1] * factorial(np.arange(order + 1))
        return np.array([fitx, fity]), dpa, np.array(qz)


def gen_detuning_elem(ring: Lattice, orbit: Optional[Orbit] = None) -> Element:
    """Generates a detuning element

    Parameters:
        ring:           Lattice description
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).

    Returns:
        detuneElem:     Element reproducing the detuning of the ring with
          amplitude and momentum
    """
    if orbit is None:
        orbit, _ = find_orbit(ring)
    lindata0, _, _ = ring.linopt6(get_chrom=False, orbit=orbit)
    xsi = get_chrom(ring.disable_6d(copy=True))
    r0, r1, x, q_dx, y, q_dy = detuning(
        ring.radiation_off(copy=True), xm=1.0e-4, ym=1.0e-4, npoints=3
    )
    nonlin_elem = Element(
        "NonLinear",
        PassMethod="DeltaQPass",
        Betax=lindata0.beta[0],
        Betay=lindata0.beta[1],
        Alphax=lindata0.alpha[0],
        Alphay=lindata0.alpha[1],
        chromx_arr=np.array([xsi[0]]),
        chromy_arr=np.array([xsi[1]]),
        A1=r1[0][0],
        A2=r1[0][1],
        A3=r1[1][1],
        T1=-orbit,
        T2=orbit,
        chrom_maxorder=1
    )
    return nonlin_elem



"""Access to properties of RF cavities

See :doc:`/p/howto/CavityControl` for a guide on using these functions.
"""

from __future__ import annotations

import functools
from enum import Enum

import numpy as np

from .elements import RFCavity
from .lattice_object import Lattice
from .utils import AtError, make_copy, Refpts

__all__ = [
    "frequency_control",
    "get_rf_frequency",
    "set_rf_frequency",
    "get_rf_voltage",
    "set_rf_voltage",
    "get_rf_timelag",
    "set_rf_timelag",
    "set_cavity",
    "Frf",
]


def frequency_control(func):
    r"""Function to be used as decorator for
    :pycode:`func(ring, *args, **kwargs)`

    If :pycode:`ring.is_6d` is :py:obj:`True` **and** *dp*, *dct* or *df*
    is specified in *kwargs*, make a copy of *ring* with a modified
    RF frequency, remove *dp*, *dct* or *df* from *kwargs* and call
    *func* with the modified *ring*.

    If :pycode:`ring.is_6d` is :py:obj:`False` **or** no *dp*, *dct* or
    *df* is specified in *kwargs*, *func* is called with *ring* unchanged.

    Example:

        >>> @frequency_control
        ... def func(ring, *args, dp=None, dct=None, df=None, **kwargs):
        ...     pass
    """

    @functools.wraps(func)
    def wrapper(ring, *args, **kwargs):
        if ring.is_6d:
            momargs = {}
            for key in ["dp", "dct", "df"]:
                v = kwargs.pop(key, None)
                if v is not None:
                    momargs[key] = v
            if len(momargs) > 0:
                ring = set_rf_frequency(ring, copy=True, **momargs)
        return func(ring, *args, **kwargs)

    return wrapper


class Frf(Enum):
    """Enum class for frequency setting"""

    #:  Constant used as a frequency value, standing for the nominal frequency
    NOMINAL = "nominal"


def _selected_cavities(ring: Lattice, cavpts: Refpts) -> list[RFCavity]:
    """Select the cavities"""
    if cavpts is None:
        try:
            cavpts = ring.cavpts
        except AttributeError:
            cavpts = RFCavity
    return cavpts


def _singlev(values, attr):
    """Return the single value"""
    vals = np.unique(values)
    if len(vals) > 1:
        raise AtError(f"{attr} not equal for all cavities")
    return vals[0]


# noinspection PyUnusedLocal
def _sumv(values, attr):
    """Return the sum of values"""
    return np.sum(values)


def _fundmask(cavities: list[RFCavity]):
    freqs = np.array([cav.Frequency for cav in cavities])
    if len(freqs) < 1:
        raise AtError("No cavity found in the lattice")
    mask = freqs == np.min(freqs)
    return mask


def _get_cavity(ring: Lattice, attr, fsingle, cavpts=None, array=False):
    cavpts = _selected_cavities(ring, cavpts)
    cavities = list(ring.select(cavpts))
    vcell = np.array([getattr(cav, attr) for cav in cavities])
    if array:
        return vcell
    else:
        fundmask = _fundmask(cavities)
        if fundmask is not None:
            vcell = vcell[fundmask]
        return fsingle(vcell, attr)


def get_rf_frequency(ring: Lattice, **kwargs) -> float:
    """Return the RF frequency

    Parameters:
        ring:               Lattice description

    Keyword Arguments:
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities. The main cavities are those with the lowest
          frequency among the selected ones.
        array (bool):       If :py:obj:`False`, return the frequency of the main
          cavities (selected cavities with the lowest frequency).

          If :py:obj:`True`, return the frequency of each selected cavity.

    Returns:
        rf_freq:    RF frequency [Hz]
    """
    return _get_cavity(ring, "Frequency", _singlev, **kwargs)


def get_rf_voltage(ring: Lattice, **kwargs) -> float:
    """Return the total RF voltage (full ring including periodicity)

    Parameters:
        ring:               Lattice description

    Keyword Arguments:
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities. The main cavities are those with the lowest
          frequency among the selected ones.
        array (bool):       If :py:obj:`False`, return the sum of the voltage of the
          main cavities (selected cavities with the lowest frequency).

          If :py:obj:`True`, return the voltage of each selected cavity.

    Returns:
        rf_v:       Total voltage (full ring including periodicity) [V]
    """
    vcell = _get_cavity(ring, "Voltage", _sumv, **kwargs)
    return ring.periodicity * vcell


def get_rf_timelag(ring: Lattice, **kwargs) -> float:
    """Return the RF time lag

    Parameters:
        ring:           Lattice description

    Keyword Arguments:
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities. The main cavities are those with the lowest
          frequency among the selected ones.
        array (bool):       If :py:obj:`False`, return the timelag of the main cavities
          (selected cavities with the lowest frequency).

          If :py:obj:`True`, return the timelag of each selected cavity.

    Returns:
        rf_timelag: RF time lag as path lengthening [m]
    """
    return _get_cavity(ring, "TimeLag", _singlev, **kwargs)


def set_rf_frequency(
    ring: Lattice, frequency: float | Frf = Frf.NOMINAL, **kwargs
) -> Lattice | None:
    """Set the RF frequency

    Parameters:
        ring:           Lattice description
        frequency:      RF frequency [Hz]. Use :py:obj:`.Frf.NOMINAL` to specify the
          nominal frequency corresponding to the given off-momentum.

    Keyword Args:
        dp (float | None):  Momentum deviation. Defaults to :py:obj:`None`
        dct (float | None): Path lengthening. Defaults to :py:obj:`None`
        df (float | None):  Deviation of RF frequency. Defaults to :py:obj:`None`
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities.
        array (bool):       If :py:obj:`False` (default), *frequency* is applied to
          the selected cavities with the lowest frequency. The frequency of all the
          other selected cavities is scaled by the same ratio.

          If :py:obj:`True`, directly apply *frequency* to the selected cavities. The
          value must be broadcastable to the number of cavities.
        copy (bool):        If :py:obj:`True`, returns a shallow copy of *ring* with
          new cavity elements. Otherwise (default), modify *ring* in-place.
    """
    return set_cavity(ring, Frequency=frequency, **kwargs)


def set_rf_voltage(ring: Lattice, voltage: float, **kwargs) -> Lattice | None:
    """Set the RF voltage for the full ring

    Parameters:
        ring:           Lattice description
        voltage:        Total RF voltage (full ring taking periodicity into account) [V]

    Keyword Arguments:
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities.
        array (bool):       If :py:obj:`False` (default), the voltage of the main
          cavities is scaled to achieve the desired *voltage*. The voltage of the other
          cavities is scaled by the same ratio.

          If :py:obj:`True`, directly apply *voltage* to the selected cavities. The
          value must be broadcastable to the number of cavities.
        copy (bool):        If :py:obj:`True`, returns a shallow copy of *ring* with
          new cavity elements. Otherwise (default), modify *ring* in-place.
    """
    return set_cavity(ring, Voltage=voltage, **kwargs)


def set_rf_timelag(ring: Lattice, timelag: float, **kwargs) -> Lattice | None:
    """Set the RF time lag

    Parameters:
        ring:           Lattice description
        timelag:        RF time lag, expressed as path lengthening [m]

    Keyword Arguments:
        cavpts (Refpts | None):  If :py:obj:`None`, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities.
        array (bool):       If :py:obj:`False` (default), *timelag* is applied to the
          main cavities. The timelag of the other cavities is shifted by the same
          amount.

          If :py:obj:`True`, directly apply *timelag* to the selected cavities. The
          value must be broadcastable to the number of cavities.
        copy (bool):    If :py:obj:`True`, returns a shallow copy of *ring* with new
          cavity elements. Otherwise (default), modify *ring* in-place.
    """
    return set_cavity(ring, TimeLag=timelag, **kwargs)


# noinspection PyPep8Naming
def set_cavity(
    ring: Lattice,
    Voltage: float | None = None,
    Frequency: float | Frf | None = None,
    TimeLag: float | None = None,
    cavpts: Refpts | None = None,
    copy: bool = False,
    array: bool = False,
    **kwargs,
) -> Lattice | None:
    """
    Set the parameters of the RF cavities

    Parameters:
        ring:           lattice description
        Frequency:      RF frequency [Hz]. Use :py:obj:`.Frf.NOMINAL` to specify the
          nominal frequency corresponding to the given off-momentum.
        Voltage:        Total RF voltage [V]
        TimeLag:        RF time shift expressed as path lengthening [m]
        cavpts:         Cavity location. If None, look for :pycode:`ring.cavpts`, or
          otherwise take all cavities
        array:          If :py:obj:`False`, the value is applied as described for
          :py:func:`set_rf_voltage`, :py:func:`set_rf_timelag` and
          :py:func:`set_rf_frequency`.

          If :py:obj:`True`, directly apply the values to the selected cavities. The
          values must be broadcastable to the number of cavities.
        copy (bool):    If :py:obj:`True`, returns a shallow copy of *ring* with new
          cavity elements. Otherwise (default), modify *ring* in-place.
    """

    # noinspection PyShadowingNames
    def getv(cavities: list[RFCavity], attr: str):
        values = np.array([getattr(el, attr) for el in cavities])
        valfund = values[fundmask]
        return values, valfund

    def nominal(elem: RFCavity, cell_frev: float) -> float:
        cell_h = getattr(elem, "HarmNumber", 0)
        if cell_h == 0:
            cell_h = round(elem.Frequency / cell_frev)
        return cell_h * cell_frev

    # noinspection PyShadowingNames
    @make_copy(copy)
    def apply(rg: Lattice, cavpts, modif):
        cavities = list(rg.select(cavpts))
        ncavs = len(cavities)
        for attr, values in modif.items():
            try:
                values = np.broadcast_to(values, (ncavs,))
            except ValueError:
                raise ValueError(
                    f"{attr} should be either scalar or a ({ncavs},) array"
                ) from None
            for val, cavity in zip(values, cavities):
                setattr(cavity, attr, val)

    cavpts = _selected_cavities(ring, cavpts)
    cavities = list(ring.select(cavpts))
    fundmask = _fundmask(cavities)

    modif = {}
    if Frequency is not None:
        if Frequency is Frf.NOMINAL:
            cell_f = ring.get_revolution_frequency(**kwargs) * ring.periodicity
            Frequency = np.array([nominal(el, cell_f) for el in cavities])
        elif not array:
            vall, vmain = getv(cavities, "Frequency")
            vmain = vmain[0]
            if vmain == 0.0:
                vall[fundmask] = Frequency
                Frequency = vall
            else:
                Frequency *= vall / vmain
        modif["Frequency"] = Frequency

    if TimeLag is not None:
        if not array:
            vall, vmain = getv(cavities, "TimeLag")
            TimeLag += vall - _singlev(vmain, "TimeLag")
        modif["TimeLag"] = TimeLag

    if Voltage is not None:
        vcell = Voltage / ring.periodicity
        if not array:
            vall, vmain = getv(cavities, "Voltage")
            vmain = np.sum(vmain)
            if vmain == 0.0:
                vall[fundmask] = vcell / np.count_nonzero(fundmask)
                vcell = vall
            else:
                vcell *= vall / vmain
        modif["Voltage"] = vcell

    return apply(ring, cavpts, modif)


Lattice.get_rf_voltage = get_rf_voltage
Lattice.get_rf_frequency = get_rf_frequency
Lattice.get_rf_timelag = get_rf_timelag
Lattice.set_rf_voltage = set_rf_voltage
Lattice.set_rf_frequency = set_rf_frequency
Lattice.set_rf_timelag = set_rf_timelag
Lattice.set_cavity = set_cavity
Lattice.rf_voltage = property(
    get_rf_voltage, set_rf_voltage, doc="RF voltage of the full ring [V]"
)
Lattice.rf_timelag = property(
    get_rf_timelag, set_rf_timelag, doc="Time lag of the fundamental mode [m]"
)
Lattice.rf_frequency = property(
    get_rf_frequency,
    set_rf_frequency,
    doc="Fundamental RF frequency [Hz]. The special value "
    ":py:class:`Frf.NOMINAL <.Frf>` means nominal frequency.",
)



# noinspection PyUnresolvedReferences
r"""Definition of :py:class:`.ResponseMatrix` objects.

A :py:class:`ResponseMatrix` object defines a general-purpose response matrix, based
on a :py:class:`.VariableList` of attributes which will be independently varied, and an
:py:class:`.ObservableList` of attributes which will be recorded for each
variable step.

:py:class:`ResponseMatrix` objects can be combined with the "+" operator to define
combined responses. This concatenates the variables and the observables.

This module also defines two commonly used response matrices:
:py:class:`OrbitResponseMatrix` for circular machines and
:py:class:`TrajectoryResponseMatrix` for beam lines. Other matrices can be easily
defined by providing the desired Observables and Variables to the
:py:class:`ResponseMatrix` base class.

Generic response matrix
-----------------------

The :py:class:`ResponseMatrix` class defines a general-purpose response matrix, based
on a :py:class:`.VariableList` of quantities which will be independently varied, and an
:py:class:`.ObservableList` of quantities which will be recorded for each step.

For instance let's take the horizontal displacements of all quadrupoles as variables:

>>> variables = VariableList(
...     RefptsVariable(ik, "dx", name=f"dx_{ik}", delta=0.0001)
...     for ik in ring.get_uint32_index(at.Quadrupole)
... )

The variables are the horizontal displacement ``dx`` of all quadrupoles. The variable
name is set to *dx_nnnn* where *nnnn* is the index of the quadrupole in the lattice.
The step is set to 0.0001 m.

Let's take the horizontal positions at all beam position monitors as observables:

>>> observables = at.ObservableList([at.OrbitObservable(at.Monitor, axis="x")])

This is a single observable named *orbit[x]* by default, with multiple values.

Instantiation
^^^^^^^^^^^^^

>>> resp_dx = at.ResponseMatrix(ring, variables, observables)

At that point, the response matrix is empty.

Matrix Building
^^^^^^^^^^^^^^^

The response matrix may be filled by several means:

#. Direct assignment of an array to the :py:attr:`~.ResponseMatrix.response` property.
   The shape of the array is checked.
#. :py:meth:`~ResponseMatrix.load` loads data from a file containing previously
   saved values or experimentally measured values,
#. :py:meth:`~ResponseMatrix.build_tracking` computes the matrix using tracking,
#. For some specialized response matrices a
   :py:meth:`~OrbitResponseMatrix.build_analytical` method is available.

Matrix normalisation
^^^^^^^^^^^^^^^^^^^^

To be correctly inverted, the response matrix must be correctly normalised: the norms
of its columns must be of the same order of magnitude, and similarly for the rows.

Normalisation is done by adjusting the weights :math:`w_v` for the variables
:math:`\mathbf{V}` and :math:`w_o` for the observables :math:`\mathbf{O}`.
With :math:`\mathbf{R}` the response matrix:

.. math::

   \mathbf{O} = \mathbf{R} . \mathbf{V}

The weighted response matrix :math:`\mathbf{R}_w` is:

.. math::

   \frac{\mathbf{O}}{w_o} = \mathbf{R}_w . \frac{\mathbf{V}}{w_v}

The :math:`\mathbf{R}_w` is dimensionless and should be normalised. This can be checked
using:

* :py:meth:`~ResponseMatrix.check_norm` which prints the ratio of the maximum / minimum
  norms for variables and observables. These should be less than 10.
* :py:meth:`~.ResponseMatrix.plot_norm`

Both natural and weighted response matrices can be retrieved with the
:py:attr:`~ResponseMatrix.response` and :py:attr:`~ResponseMatrix.weighted_response`
properties.

Matrix pseudo-inversion
^^^^^^^^^^^^^^^^^^^^^^^

The :py:meth:`~ResponseMatrix.solve` method computes the singular values of the
weighted response matrix.

After solving, correction is available, for instance with

* :py:meth:`~ResponseMatrix.correction_matrix` which returns the correction matrix
  (pseudo-inverse of the response matrix),
* :py:meth:`~ResponseMatrix.get_correction` which returns a correction vector when
  given error values,
* :py:meth:`~ResponseMatrix.correct` which computes and optionally applies a correction
  for the provided :py:class:`.Lattice`.

Exclusion of variables and observables
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Variables may be added to a set of excluded values, and similarly for observables.
Excluding an item does not change the response matrix. The values are excluded from the
pseudo-inversion of the response, possibly reducing the number of singular values.
After inversion the correction matrix is expanded to its original size by inserting
zero lines and columns at the location of excluded items. This way:

- error and correction vectors keep the same size independently of excluded values,
- excluded error values are ignored,
- excluded corrections are set to zero.

Variables can be added to the set of excluded variables using
:py:meth:`~.ResponseMatrix.exclude_vars` and observables using
:py:meth:`~.ResponseMatrix.exclude_obs`.

After excluding items, the pseudo-inverse is discarded so one must recompute it again
by calling :py:meth:`~ResponseMatrix.solve`.

The exclusion masks can be reset with  :py:meth:`~.ResponseMatrix.reset_vars` and
:py:meth:`~.ResponseMatrix.reset_obs`.
"""

from __future__ import annotations

__all__ = [
    "sequence_split",
    "ResponseMatrix",
    "OrbitResponseMatrix",
    "TrajectoryResponseMatrix",
]

import os
import multiprocessing
import concurrent.futures
import abc
import warnings
from collections.abc import Sequence, Generator, Callable
from typing import Any, ClassVar
from itertools import chain
from functools import partial
import math

import numpy as np
import numpy.typing as npt

from .observables import ElementObservable
from .observables import TrajectoryObservable, OrbitObservable, LatticeObservable
from .observables import LocalOpticsObservable, GlobalOpticsObservable
from .observablelist import ObservableList
from ..lattice import AtError, AtWarning, Refpts, Uint32Refpts, All
from ..lattice import AxisDef, plane_, Lattice, Monitor, checkattr
from ..lattice.lattice_variables import RefptsVariable
from ..lattice.variables import VariableList

FloatArray = npt.NDArray[np.float64]

_orbit_correctors = checkattr("KickAngle")

_globring: Lattice | None = None
_globobs: ObservableList | None = None

warnings.filterwarnings("always", category=AtWarning, module=__name__)


def sequence_split(seq: Sequence, nslices: int) -> Generator[Sequence, None, None]:
    """Split a sequence into multiple sub-sequences.

    The length of *seq* does not have to be a multiple of *nslices*.

    Args:
        seq: sequence to split
        nslices: number of sub-sequences

    Returns:
        subseqs: Iterator over sub-sequences
    """

    def _split(seqsizes):
        beg = 0
        for size in seqsizes:
            end = beg + size
            yield seq[beg:end]
            beg = end

    lna = len(seq)
    sz, rem = divmod(lna, nslices)
    lsubseqs = [sz] * nslices
    for k in range(rem):
        lsubseqs[k] += 1
    return _split(lsubseqs)


def _resp(
    ring: Lattice, observables: ObservableList, variables: VariableList, **kwargs
):
    def _resp_one(variable: RefptsVariable):
        """Single response"""
        variable.step_up(ring=ring)
        observables.evaluate(ring, **kwargs)
        op = observables.flat_values
        variable.step_down(ring=ring)
        observables.evaluate(ring, **kwargs)
        om = observables.flat_values
        variable.reset(ring=ring)
        return (op - om) / (2.0 * variable.delta)

    return [_resp_one(v) for v in variables]


def _resp_fork(variables: VariableList, **kwargs):
    """Response for fork parallel method."""
    return _resp(_globring, _globobs, variables, **kwargs)


class _SvdSolver(abc.ABC):
    """SVD solver for response matrices."""

    _shape: tuple[int, int]
    _obsmask: npt.NDArray[bool]
    _varmask: npt.NDArray[bool]
    _response: FloatArray | None = None
    _v: FloatArray | None = None
    _uh: FloatArray | None = None
    #: Singular values of the response matrix
    singular_values: FloatArray | None = None

    def __init__(self, nobs: int, nvar: int):
        self._shape = (nobs, nvar)
        self._obsmask = np.ones(nobs, dtype=bool)
        self._varmask = np.ones(nvar, dtype=bool)

    def reset_vars(self):
        """Reset the variable exclusion mask: enable all variables"""
        self._varmask = np.ones(self.shape[1], dtype=bool)
        self._v = None
        self._uh = None
        self.singular_values = None

    def reset_obs(self):
        """Reset the observable exclusion mask: enable all observables"""
        self._obsmask = np.ones(self.shape[0], dtype=bool)
        self._v = None
        self._uh = None
        self.singular_values = None

    @property
    @abc.abstractmethod
    def varweights(self) -> np.ndarray: ...

    @property
    @abc.abstractmethod
    def obsweights(self) -> np.ndarray: ...

    @property
    def shape(self) -> tuple[int, int]:
        """Shape of the response matrix."""
        return self._shape

    def solve(self) -> None:
        """Compute the singular values of the response matrix."""
        resp = self.weighted_response
        selected = np.ix_(self._obsmask, self._varmask)
        u, s, vh = np.linalg.svd(resp[selected], full_matrices=False)
        self._v = vh.T * (1.0 / s) * self.varweights[self._varmask].reshape(-1, 1)
        self._uh = u.T / self.obsweights[self._obsmask]
        self.singular_values = s

    def check_norm(self) -> tuple[FloatArray, FloatArray]:
        """Display the norm of the rows and columns of the weighted response matrix.

        Adjusting the variables and observable weights to equalize the norms
        of rows and columns is important.

        Returns:
            obs_norms:      Norms of observables (rows)
            var_norms:      Norms of Variables (columns)
        """
        resp = self.weighted_response
        obs = np.linalg.norm(resp, axis=1)
        var = np.linalg.norm(resp, axis=0)
        print(f"max/min Observables: {np.amax(obs) / np.amin(obs)}")
        print(f"max/min Variables: {np.amax(var) / np.amin(var)}")
        return obs, var

    @property
    def response(self) -> FloatArray:
        """Response matrix."""
        resp = self._response
        if resp is None:
            raise AtError("No matrix yet: run build() or load() first")
        return resp

    @response.setter
    def response(self, response: FloatArray) -> None:
        l1, c1 = self._shape
        l2, c2 = response.shape
        if l1 != l1 or c1 != c2:
            raise ValueError(
                f"Input matrix has incompatible shape. Expected: {self.shape}"
            )
        self._response = response

    @property
    def weighted_response(self) -> FloatArray:
        """Weighted response matrix."""
        return self.response * (self.varweights / self.obsweights.reshape(-1, 1))

    def correction_matrix(self, nvals: int | None = None) -> FloatArray:
        """Return the correction matrix (pseudo-inverse of the response matrix).

        Args:
            nvals:  Desired number of singular values. If :py:obj:`None`, use
              all singular values

        Returns:
            cormat: Correction matrix
        """
        if self.singular_values is None:
            self.solve()
        if nvals is None:
            nvals = len(self.singular_values)
        cormat = np.zeros(self._shape[::-1])
        selected = np.ix_(self._varmask, self._obsmask)
        cormat[selected] = self._v[:, :nvals] @ self._uh[:nvals, :]
        return cormat

    def get_correction(
        self, observed: FloatArray, nvals: int | None = None
    ) -> FloatArray:
        """Compute the correction of the given observation.

        Args:
            observed:   Vector of observed deviations,
            nvals:      Desired number of singular values. If :py:obj:`None`, use
              all singular values

        Returns:
            corr:       Correction vector
        """
        return -self.correction_matrix(nvals=nvals) @ observed

    def save(self, file) -> None:
        """Save a response matrix in the NumPy .npy format.

        Args:
            file:   file-like object, string, or :py:class:`pathlib.Path`: File to
              which the data is saved. If file is a file-object, it must be opened in
              binary mode. If file is a string or Path, a .npy extension will
              be appended to the filename if it does not already have one.
        """
        if self._response is None:
            raise AtError("No response matrix: run build_tracking() or load() first")
        np.save(file, self._response)

    def load(self, file) -> None:
        """Load a response matrix saved in the NumPy .npy format.

        Args:
            file:   file-like object, string, or :py:class:`pathlib.Path`: the file to
              read. A file object must always be opened in binary mode.
        """
        self.response = np.load(file)


class ResponseMatrix(_SvdSolver):
    r"""Base class for response matrices.

    It is defined by any arbitrary set of :py:class:`~.variables.VariableBase`\ s and
    :py:class:`.Observable`\s

    Addition is defined on :py:class:`ResponseMatrix` objects as the addition
    of their :py:class:`~.variables.VariableBase`\ s and :py:class:`.Observable`\s to
    produce combined responses.
    """

    ring: Lattice
    variables: VariableList  #: List of matrix :py:class:`Variable <.VariableBase>`\ s
    observables: ObservableList  #: List of matrix :py:class:`.Observable`\s
    _eval_args: dict[str, Any] = {}

    def __init__(
        self,
        ring: Lattice,
        variables: VariableList,
        observables: ObservableList,
    ):
        r"""
        Args:
            ring:           Design lattice, used to compute the response
            variables:      List of :py:class:`Variable <.VariableBase>`\ s
            observables:    List of :py:class:`.Observable`\s
        """

        def limits(obslist):
            beg = 0
            for obs in obslist:
                end = beg + obs.value.size
                yield beg, end
                beg = end

        # for efficiency of parallel computation, the variable's refpts must be integer
        for var in variables:
            var.refpts = ring.get_uint32_index(var.refpts)
        self.ring = ring
        self.variables = variables
        self.observables = observables
        variables.get(ring=ring, initial=True)
        observables.evaluate(ring=ring, initial=True)
        super().__init__(len(observables.flat_values), len(variables))
        self._ob = [self._obsmask[beg:end] for beg, end in limits(self.observables)]

    def __add__(self, other: ResponseMatrix):
        if not isinstance(other, ResponseMatrix):
            raise TypeError(
                f"Cannot add {type(other).__name__} and {type(self).__name__}"
            )
        return ResponseMatrix(
            self.ring,
            VariableList(self.variables + other.variables),
            self.observables + other.observables,
        )

    def __str__(self):
        no, nv = self.shape
        return f"{type(self).__name__}({no} observables, {nv} variables)"

    @property
    def varweights(self) -> np.ndarray:
        """Variable weights."""
        return self.variables.deltas

    @property
    def obsweights(self) -> np.ndarray:
        """Observable weights."""
        return self.observables.flat_weights

    def correct(
        self, ring: Lattice, nvals: int = None, niter: int = 1, apply: bool = False
    ) -> FloatArray:
        """Compute and optionally apply the correction.

        Args:
            ring:       Lattice description. The response matrix observables
              will be evaluated for *ring* and the deviation from target will
              be corrected
            apply:      If :py:obj:`True`, apply the correction to *ring*
            niter:      Number of iterations. For more than one iteration,
              *apply* must be :py:obj:`True`
            nvals:      Desired number of singular values. If :py:obj:`None`,
              use all singular values. *nvals* may be a scalar or an iterable with
              *niter* values.

        Returns:
            correction: Vector of correction values
        """
        if niter > 1 and not apply:
            raise ValueError("If niter > 1, 'apply' must be True")
        obs = self.observables
        if apply:
            self.variables.get(ring=ring, initial=True)
        sumcorr = np.array([0.0])
        for it, nv in zip(range(niter), np.broadcast_to(nvals, (niter,))):
            print(f'step {it+1}, nvals = {nv}')
            obs.evaluate(ring, **self._eval_args)
            err = obs.flat_deviations
            if np.any(np.isnan(err)):
                raise AtError(
                    f"Step {it + 1}: Invalid observables, cannot compute correction"
                )
            corr = self.get_correction(obs.flat_deviations, nvals=nv)
            sumcorr = sumcorr + corr  # non-broadcastable sumcorr
            if apply:
                self.variables.increment(corr, ring=ring)
        return sumcorr

    def build_tracking(
        self,
        use_mp: bool = False,
        pool_size: int | None = None,
        start_method: str | None = None,
        **kwargs,
    ) -> FloatArray:
        """Build the response matrix.

        Args:
            use_mp:             Use multiprocessing
            pool_size:          number of processes. If None,
              :pycode:`min(len(self.variables, nproc)` is used
            start_method:       python multiprocessing start method.
              :py:obj:`None` uses the python default that is considered safe.
              Available values: ``'fork'``, ``'spawn'``, ``'forkserver'``.
              Default for linux is ``'fork'``, default for macOS and  Windows
              is ``'spawn'``. ``'fork'`` may be used on macOS to speed up the
              calculation, however it is considered unsafe.

        Keyword Args:
            dp (float):     Momentum deviation. Defaults to :py:obj:`None`
            dct (float):    Path lengthening. Defaults to :py:obj:`None`
            df (float):     Deviation from the nominal RF frequency.
              Defaults to :py:obj:`None`
            r_in (Orbit):   Initial trajectory, used for
              :py:class:`TrajectoryResponseMatrix`, Default: zeros(6)

        Returns:
            response:       Response matrix
        """
        self._eval_args = kwargs
        self.observables.evaluate(self.ring)
        ring = self.ring.deepcopy()

        if use_mp:
            global _globring
            global _globobs
            ctx = multiprocessing.get_context(start_method)
            if pool_size is None:
                pool_size = min(len(self.variables), os.cpu_count())
            obschunks = sequence_split(self.variables, pool_size)
            if ctx.get_start_method() == "fork":
                _globring = ring
                _globobs = self.observables
                _single_resp = partial(_resp_fork, **kwargs)
            else:
                _single_resp = partial(_resp, ring, self.observables, **kwargs)
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=pool_size,
                mp_context=ctx,
            ) as pool:
                results = list(chain(*pool.map(_single_resp, obschunks)))
            _globring = None
            _globobs = None
        else:
            results = _resp(ring, self.observables, self.variables, **kwargs)

        resp = np.stack(results, axis=-1)
        self.response = resp
        return resp

    def build_analytical(self) -> FloatArray:
        """Build the response matrix."""
        raise NotImplementedError(
            f"build_analytical not implemented for {self.__class__.__name__}"
        )

    def _on_obs(self, fun: Callable, *args, obsid: int | str = 0):
        """Apply a function to the selected observable"""
        if not isinstance(obsid, str):
            return fun(self.observables[obsid], *args)
        else:
            for obs in self.observables:
                if obs.name == obsid:
                    return fun(obs, *args)
            else:
                raise ValueError(f"Observable {obsid} not found")

    def get_target(self, *, obsid: int | str = 0) -> FloatArray:
        r"""Return the target of the specified observable

        Args:
            obsid:  :py:class:`.Observable` name or index in the observable list.

        Returns:
            target: observable target
        """
        def _get(obs):
            return obs.target

        return self._on_obs(_get, obsid=obsid)

    def set_target(self, target: npt.ArrayLike, *, obsid: int | str = 0) -> None:
        r"""Set the target of the specified observable

        Args:
            target: observable target. Must be broadcastable to the shape of the
              observable value.
            obsid:  :py:class:`.Observable` name or index in the observable list.
        """

        def _set(obs, targ):
            obs.target = targ

        return self._on_obs(_set, target, obsid=obsid)

    def exclude_obs(self, *, obsid: int | str = 0, refpts: Refpts = None) -> None:
        # noinspection PyUnresolvedReferences
        r"""Add an observable item to the set of excluded values

        After excluding observation points, the matrix must be inverted again using
        :py:meth:`solve`.

        Args:
            obsid:      :py:class:`.Observable` name or index in the observable list.
            refpts:     location of elements to exclude for
              :py:class:`.ElementObservable` objects, otherwise ignored.

        Raises:
            ValueError: No observable with the given name.
            IndexError: Observableindex out of range.

        Example:
            >>> resp = OrbitResponseMatrix(ring, "h", Monitor, Corrector)
            >>> resp.exclude_obs(obsid="x_orbit", refpts="BPM_02")

            Create an horizontal :py:class:`OrbitResponseMatrix` from
            :py:class:`.Corrector` elements to :py:class:`.Monitor` elements,
            and exclude the monitor with name "BPM_02"
        """

        def exclude(ob, msk):
            inimask = msk.copy()
            if isinstance(ob, ElementObservable) and not ob.summary:
                boolref = self.ring.get_bool_index(refpts)
                # noinspection PyProtectedMember
                msk &= np.logical_not(boolref[ob._boolrefs])
            else:
                msk[:] = False
            if np.all(msk == inimask):
                warnings.warn(AtWarning("No new excluded value"), stacklevel=3)
            # Force a new computation
            self.singular_values = None

        if not isinstance(obsid, str):
            exclude(self.observables[obsid], self._ob[obsid])
        else:
            for obs, mask in zip(self.observables, self._ob):
                if obs.name == obsid:
                    exclude(obs, mask)
                    break
            else:
                raise ValueError(f"Observable {obsid} not found")

    @property
    def excluded_obs(self) -> dict:
        """Directory of excluded observables.

        The dictionary keys are the observable names, the values are the integer
        indices of excluded items (empty list if no exclusion).
        """

        def ex(obs, mask):
            if isinstance(obs, ElementObservable) and not obs.summary:
                refpts = self.ring.get_bool_index(None)
                # noinspection PyProtectedMember
                refpts[obs._boolrefs] = np.logical_not(mask)
                refpts = self.ring.get_uint32_index(refpts)
            else:
                refpts = np.arange(0 if np.all(mask) else mask.size, dtype=np.uint32)
            return refpts

        return {ob.name: ex(ob, mask) for ob, mask in zip(self.observables, self._ob)}

    def exclude_vars(self, *varid: int | str) -> None:
        # noinspection PyUnresolvedReferences
        """Add variables to the set of excluded variables.

        Args:
            *varid:  :py:class:`Variable <.VariableBase>` names or variable indices
              in the variable list

        After excluding variables, the matrix must be inverted again using
        :py:meth:`solve`.

        Examples:
            >>> resp.exclude_vars(0, "var1", -1)

            Exclude the 1st variable, the variable named "var1" and the last variable.
        """
        nameset = set(nm for nm in varid if isinstance(nm, str))
        varidx = [nm for nm in varid if not isinstance(nm, str)]
        mask = np.array([var.name in nameset for var in self.variables])
        mask[varidx] = True
        miss = nameset - {var.name for var, ok in zip(self.variables, mask) if ok}
        if miss:
            raise ValueError(f"Unknown variables: {miss}")
        self._varmask &= np.logical_not(mask)

    @property
    def excluded_vars(self) -> list:
        """List of excluded variables"""
        return [var.name for var, ok in zip(self.variables, self._varmask) if not ok]


class OrbitResponseMatrix(ResponseMatrix):
    # noinspection PyUnresolvedReferences
    r"""Orbit response matrix.

    An :py:class:`OrbitResponseMatrix` applies to a single plane, horizontal or
    vertical. A combined response matrix is obtained by adding horizontal and
    vertical matrices. However, the resulting matrix has the :py:class:`ResponseMatrix`
    class, which implies that the :py:class:`OrbitResponseMatrix` specific methods are
    not available.

    Variables are a set of steerers and optionally the RF frequency. Steerer
    variables are named ``xnnnn`` or ``ynnnn`` where nnnn is the index in the
    lattice. The RF frequency variable is named ``RF frequency``.

    Observables are the closed orbit position at selected points, named
    ``orbit[x]`` for the horizontal plane or ``orbit[y]`` for the vertical plane,
    and optionally the sum of steerer angles named ``sum(h_kicks)`` or
    ``sum(v_kicks)``

    The variable elements must have the *KickAngle* attribute used for correction.
    It's available for all magnets, though not present by default
    except in :py:class:`.Corrector` magnets. For other magnets, the attribute
    should be explicitly created.

    By default, the observables are all the :py:class:`.Monitor` elements, and the
    variables are all the elements having a *KickAngle* attribute.
    This is equivalent to:

    >>> resp_v = OrbitResponseMatrix(
    ...     ring, "v", bpmrefs=at.Monitor, steerrefs=at.checkattr("KickAngle")
    ... )
    """

    bpmrefs: Uint32Refpts  #: location of position monitors
    steerrefs: Uint32Refpts  #: location of steerers

    def __init__(
        self,
        ring: Lattice,
        plane: AxisDef,
        bpmrefs: Refpts = Monitor,
        steerrefs: Refpts = _orbit_correctors,
        *,
        cavrefs: Refpts = None,
        bpmweight: float | Sequence[float] = 1.0,
        bpmtarget: float | Sequence[float] = 0.0,
        steerdelta: float | Sequence[float] = 0.0001,
        cavdelta: float | None = None,
        steersum: bool = False,
        stsumweight: float | None = None,
    ):
        """
        Args:
            ring:       Design lattice, used to compute the response.
            plane:      One out of {0, 'x', 'h', 'H'} for horizontal orbit, or
              one of {1, 'y', 'v', 'V'} for vertical orbit.
            bpmrefs:    Location of closed orbit observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`".
              Default: all :py:class:`.Monitor` elements.
            steerrefs:  Location of orbit steerers. Their *KickAngle* attribute
              is used and must be present in the selected elements.
              Default: All Elements having a *KickAngle* attribute.
            cavrefs:    Location of RF cavities. Their *Frequency* attribute
              is used. If :py:obj:`None`, no cavity is included in the response.
              Cavities must be active. Cavity variables are appended to the steerer
              variables.
            bpmweight:  Weight of position readings. Must be broadcastable to the
              number of BPMs.
            bpmtarget:  Target orbit position. Must be broadcastable to the number of
              observation points.
            cavdelta:   Step on RF frequency for matrix computation [Hz]. This
              is also the cavity weight. Default: automatically computed.
            steerdelta: Step on steerers for matrix computation [rad]. This is
              also the steerer weight. Must be broadcastable to the number of steerers.
            steersum:   If :py:obj:`True`, the sum of steerers is appended to the
              Observables.
            stsumweight: Weight on steerer summation. Default: automatically computed.

        :ivar VariableList variables: matrix variables
        :ivar ObservableList observables: matrix observables

        By default, the weights of cavities and steerers summation are set to give
        a factor 2 more efficiency than steerers and BPMs

        """

        def steerer(ik, delta):
            name = f"{plcode}{ik:04}"
            return RefptsVariable(ik, "KickAngle", index=pl, name=name, delta=delta)

        def set_norm():
            bpm = LocalOpticsObservable(bpmrefs, "beta", plane=pl)
            sts = LocalOpticsObservable(steerrefs, "beta", plane=pl)
            dsp = LocalOpticsObservable(bpmrefs, "dispersion", plane=2 * pl)
            tun = GlobalOpticsObservable("tune", plane=pl)
            obs = ObservableList([bpm, sts, dsp, tun])
            result = obs.evaluate(ring=ring)
            alpha = ring.disable_6d(copy=True).get_mcf(0)
            freq = ring.get_rf_frequency(cavpts=cavrefs)
            nr = np.outer(
                np.sqrt(result[0]) / bpmweight, np.sqrt(result[1]) * steerdelta
            )
            vv = np.mean(np.linalg.norm(nr, axis=0))
            vo = np.mean(np.linalg.norm(nr, axis=1))
            korb = 0.25 * math.sqrt(2.0) / math.sin(math.pi * result[3])
            cd = vv * korb * alpha * freq / np.linalg.norm(result[2] / bpmweight)
            sw = np.linalg.norm(deltas) / vo / korb
            return cd, sw

        pl = plane_(plane, key="index")
        plcode = plane_(plane, key="code")
        ids = ring.get_uint32_index(steerrefs)
        nbsteers = len(ids)
        deltas = np.broadcast_to(steerdelta, nbsteers)
        if steersum and stsumweight is None or cavrefs and cavdelta is None:
            cavd, stsw = set_norm()

        # Observables
        bpms = OrbitObservable(bpmrefs, axis=2 * pl, target=bpmtarget, weight=bpmweight)
        observables = ObservableList([bpms])
        if steersum:
            # noinspection PyUnboundLocalVariable
            sumobs = LatticeObservable(
                steerrefs,
                "KickAngle",
                name=f"{plcode}_kicks",
                target=0.0,
                index=pl,
                weight=stsumweight if stsumweight else stsw / 2.0,
                statfun=np.sum,
            )
            observables.append(sumobs)

        # Variables
        variables = VariableList(steerer(ik, delta) for ik, delta in zip(ids, deltas))
        if cavrefs is not None:
            active = (el.longt_motion for el in ring.select(cavrefs))
            if not all(active):
                raise ValueError("Cavities are not active")
            # noinspection PyUnboundLocalVariable
            cavvar = RefptsVariable(
                cavrefs,
                "Frequency",
                name="RF frequency",
                delta=cavdelta if cavdelta else 2.0 * cavd,
            )
            variables.append(cavvar)

        super().__init__(ring, variables, observables)
        self.plane = pl
        self.steerrefs = ids
        self.nbsteers = nbsteers
        self.bpmrefs = ring.get_uint32_index(bpmrefs)

    def exclude_obs(self, *, obsid: int | str = 0, refpts: Refpts = None) -> None:
        # noinspection PyUnresolvedReferences
        r"""Add an observable item to the set of excluded values.

        After excluding observation points, the matrix must be inverted again using
        :py:meth:`solve`.

        Args:
            obsid:      If 0 (default), act on Monitors. Otherwise,
              it must be 1 or "sum(x_kicks)"  or "sum(y_kicks)"
            refpts:    location of Monitors to exclude

        Raises:
            ValueError: No observable with the given name.
            IndexError: Observableindex out of range.

        Example:
            >>> resp = OrbitResponseMatrix(ring, "h")
            >>> resp.exclude_obs("BPM_02")

            Create an horizontal :py:class:`OrbitResponseMatrix` from
            :py:class:`.Corrector` elements to :py:class:`.Monitor` elements,
            and exclude all monitors with name "BPM_02"
        """
        super().exclude_obs(obsid=obsid, refpts=refpts)

    def exclude_vars(self, *varid: int | str, refpts: Refpts = None) -> None:
        # noinspection PyUnresolvedReferences
        """Add correctors to the set of excluded variables.

        Args:
            *varid:  :py:class:`Variable <.VariableBase>` names or variable indices
              in the variable list
            refpts:  location of correctors to exclude

        After excluding correctors, the matrix must be inverted again using
        :py:meth:`solve`.

        Examples:
            >>> resp.exclude_vars(0, "x0097", -1)

            Exclude the 1st variable, the variable named "x0097" and the last variable.

            >>> resp.exclude_vars(refpts="SD1E")

            Exclude all variables associated with the element named "SD1E".
        """
        plcode = plane_(self.plane, key="code")
        names = [f"{plcode}{ik:04}" for ik in self.ring.get_uint32_index(refpts)]
        super().exclude_vars(*varid, *names)

    def normalise(
        self, cav_ampl: float | None = 2.0, stsum_ampl: float | None = 2.0
    ) -> None:
        """Normalise the response matrix

        Adjust the RF cavity delta and/or the weight of steerer summation so that the
        weighted response matrix is normalised.

        Args:
            cav_ampl: Desired ratio between the cavity response and the average of
              steerer responses. If :py:obj:`None`, do not normalise.
            stsum_ampl: Desired inverse ratio between the weight of the steerer
              summation and the average of Monitor responses. If :py:obj:`None`,
              do not normalise.

        By default, the normalisation gives to the RF frequency and steerer summation
        a factor 2 more efficiency than steerers and BPMs
        """
        resp = self.weighted_response
        normvar = np.linalg.norm(resp, axis=0)
        normobs = np.linalg.norm(resp, axis=1)
        if len(self.variables) > self.nbsteers and cav_ampl is not None:
            self.cavdelta *= np.mean(normvar[:-1]) / normvar[-1] * cav_ampl
        if len(self.observables) > 1 and stsum_ampl is not None:
            self.stsumweight = (
                self.stsumweight * normobs[-1] / np.mean(normobs[:-1]) / stsum_ampl
            )

    def build_analytical(self, **kwargs) -> FloatArray:
        """Build analytically the response matrix.

        Keyword Args:
            dp (float):     Momentum deviation. Defaults to :py:obj:`None`
            dct (float):    Path lengthening. Defaults to :py:obj:`None`
            df (float):     Deviation from the nominal RF frequency.
              Defaults to :py:obj:`None`

        Returns:
            response:       Response matrix

        References:
            .. [#Franchi] A. Franchi, S.M. Liuzzo, Z. Marti, *"Analytic formulas for
               the rapid evaluation of the orbit response matrix and chromatic functions
               from lattice parameters in circular accelerators"*,
               arXiv:1711.06589 [physics.acc-ph]
        """

        def tauwj(muj, muw):
            tau = muj - muw
            if tau < 0.0:
                tau += 2.0 * pi_tune
            return tau - pi_tune

        ring = self.ring
        pl = self.plane
        _, ringdata, elemdata = ring.linopt6(All, **kwargs)
        pi_tune = math.pi * ringdata.tune[pl]
        dataw = elemdata[self.steerrefs]
        dataj = elemdata[self.bpmrefs]
        dispj = dataj.dispersion[:, 2 * pl]
        dispw = dataw.dispersion[:, 2 * pl]
        lw = np.array([elem.Length for elem in ring.select(self.steerrefs)])
        taufunc = np.frompyfunc(tauwj, 2, 1)

        sqbetaw = np.sqrt(dataw.beta[:, pl])
        ts = lw / sqbetaw / 2.0
        tc = sqbetaw - dataw.alpha[:, pl] * ts
        twj = np.astype(taufunc.outer(dataj.mu[:, pl], dataw.mu[:, pl]), np.float64)
        jcwj = tc * np.cos(twj) + ts * np.sin(twj)
        coefj = np.sqrt(dataj.beta[:, pl]) / (2.0 * np.sin(pi_tune))
        resp = coefj[:, np.newaxis] * jcwj
        if ring.is_6d:
            alpha_c = ring.disable_6d(copy=True).get_mcf()
            resp += np.outer(dispj, dispw) / (alpha_c * ring.circumference)
            if len(self.variables) > self.nbsteers:
                rfrsp = -dispj / (alpha_c * ring.rf_frequency)
                resp = np.concatenate((resp, rfrsp[:, np.newaxis]), axis=1)
        if len(self.observables) > 1:
            sumst = np.ones(resp.shape[1], np.float64)
            if len(self.variables) > self.nbsteers:
                sumst[-1] = 0.0
            resp = np.concatenate((resp, sumst[np.newaxis]), axis=0)
        self.response = resp
        return resp

    @property
    def bpmweight(self) -> FloatArray:
        """Weight of position readings."""
        return self.observables[0].weight

    @bpmweight.setter
    def bpmweight(self, value: npt.ArrayLike):
        self.observables[0].weight = value

    @property
    def stsumweight(self) -> FloatArray:
        """Weight of steerer summation."""
        return self.observables[1].weight

    @stsumweight.setter
    def stsumweight(self, value: float):
        self.observables[1].weight = value

    @property
    def steerdelta(self) -> FloatArray:
        """Step and weight of steerers."""
        return self.variables[: self.nbsteers].deltas

    @steerdelta.setter
    def steerdelta(self, value: npt.ArrayLike):
        self.variables[: self.nbsteers].deltas = value

    @property
    def cavdelta(self) -> FloatArray:
        """Step and weight of RF frequency deviation."""
        return self.variables[self.nbsteers].delta

    @cavdelta.setter
    def cavdelta(self, value: float):
        self.variables[self.nbsteers].delta = value


class TrajectoryResponseMatrix(ResponseMatrix):
    """Trajectory response matrix.

    A :py:class:`TrajectoryResponseMatrix` applies to a single plane, horizontal or
    vertical. A combined response matrix is obtained by adding horizontal and vertical
    matrices. However, the resulting matrix has the :py:class:`ResponseMatrix`
    class, which implies that the :py:class:`OrbitResponseMatrix` specific methods are
    not available.

    Variables are a set of steerers. Steerer variables are named ``xnnnn`` or
    ``ynnnn`` where *nnnn* is the index in the lattice.

    Observables are the trajectory position at selected points, named ``trajectory[x]``
    for the horizontal plane or ``trajectory[y]`` for the vertical plane.

    The variable elements must have the *KickAngle* attribute used for correction.
    It's available for all magnets, though not present by default
    except in :py:class:`.Corrector` magnets. For other magnets, the attribute
    should be explicitly created.

    By default, the observables are all the :py:class:`.Monitor` elements, and the
    variables are all the elements having a *KickAngle* attribute.

    """

    bpmrefs: Uint32Refpts
    steerrefs: Uint32Refpts
    _default_twiss_in: ClassVar[dict] = {"beta": np.ones(2), "alpha": np.zeros(2)}

    def __init__(
        self,
        ring: Lattice,
        plane: AxisDef,
        bpmrefs: Refpts = Monitor,
        steerrefs: Refpts = _orbit_correctors,
        *,
        bpmweight: float = 1.0,
        bpmtarget: float = 0.0,
        steerdelta: float = 0.0001,
    ):
        """
        Args:
            ring:       Design lattice, used to compute the response
            plane:      One out of {0, 'x', 'h', 'H'} for horizontal orbit, or
              one of {1, 'y', 'v', 'V'} for vertical orbit
            bpmrefs:    Location of closed orbit observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`".
              Default: all :py:class:`.Monitor` elements.
            steerrefs:  Location of orbit steerers. Their *KickAngle* attribute
              is used and must be present in the selected elements.
              Default: All Elements having a *KickAngle* attribute.
            bpmweight:  Weight on position readings. Must be broadcastable to the
              number of BPMs
            bpmtarget:  Target position
            steerdelta: Step on steerers for matrix computation [rad]. This is
              also the steerer weight. Must be broadcastable to the number of steerers.
        """

        def steerer(ik, delta):
            name = f"{plcode}{ik:04}"
            return RefptsVariable(ik, "KickAngle", index=pl, name=name, delta=delta)

        pl = plane_(plane, key="index")
        plcode = plane_(plane, key="code")
        ids = ring.get_uint32_index(steerrefs)
        nbsteers = len(ids)
        deltas = np.broadcast_to(steerdelta, nbsteers)
        # Observables
        bpms = TrajectoryObservable(
            bpmrefs, axis=2 * pl, target=bpmtarget, weight=bpmweight
        )
        observables = ObservableList([bpms])
        # Variables
        variables = VariableList(steerer(ik, delta) for ik, delta in zip(ids, deltas))

        super().__init__(ring, variables, observables)
        self.plane = pl
        self.steerrefs = ids
        self.nbsteers = nbsteers
        self.bpmrefs = ring.get_uint32_index(bpmrefs)

    def build_analytical(self, **kwargs) -> FloatArray:
        """Build analytically the response matrix.

        Keyword Args:
            dp (float):     Momentum deviation. Defaults to :py:obj:`None`
            dct (float):    Path lengthening. Defaults to :py:obj:`None`
            df (float):     Deviation from the nominal RF frequency.
              Defaults to :py:obj:`None`

        Returns:
            response:       Response matrix
        """
        ring = self.ring
        pl = self.plane
        twiss_in = self._eval_args.get("twiss_in", self._default_twiss_in)
        _, _, elemdata = ring.linopt6(All, twiss_in=twiss_in, **kwargs)
        dataj = elemdata[self.bpmrefs]
        dataw = elemdata[self.steerrefs]
        lw = np.array([elem.Length for elem in ring.select(self.steerrefs)])

        sqbetaw = np.sqrt(dataw.beta[:, pl])
        ts = lw / sqbetaw / 2.0
        tc = sqbetaw - dataw.alpha[:, pl] * ts
        twj = dataj.mu[:, pl].reshape(-1, 1) - dataw.mu[:, pl]
        jswj = tc * np.sin(twj) - ts * np.cos(twj)
        coefj = np.sqrt(dataj.beta[:, pl])
        resp = coefj[:, np.newaxis] * jswj
        resp[twj < 0.0] = 0.0
        self.response = resp
        return resp

    def exclude_obs(self, *, obsid: int | str = 0, refpts: Refpts = None) -> None:
        # noinspection PyUnresolvedReferences
        r"""Add a monitor to the set of excluded values.

        After excluding observation points, the matrix must be inverted again using
        :py:meth:`solve`.

        Args:
            refpts:    location of Monitors to exclude

        Raises:
            ValueError: No observable with the given name.
            IndexError: Observableindex out of range.

        Example:
            >>> resp = TrajectoryResponseMatrix(ring, "v")
            >>> resp.exclude_obs("BPM_02")

            Create a vertical :py:class:`TrajectoryResponseMatrix` from
            :py:class:`.Corrector` elements to :py:class:`.Monitor` elements,
            and exclude all monitors with name "BPM_02"
        """
        super().exclude_obs(obsid=0, refpts=refpts)

    def exclude_vars(self, *varid: int | str, refpts: Refpts = None) -> None:
        # noinspection PyUnresolvedReferences
        """Add correctors to the set of excluded variables.

        Args:
            *varid:  :py:class:`Variable <.VariableBase>` names or variable indices
              in the variable list
            refpts:  location of correctors to exclude

        After excluding correctors, the matrix must be inverted again using
        :py:meth:`solve`.

        Examples:
            >>> resp.exclude_vars(0, "x0103", -1)

            Exclude the 1st variable, the variable named "x0103" and the last variable.

            >>> resp.exclude_vars(refpts="SD1E")

            Exclude all variables associated with the element named "SD1E".
        """
        plcode = plane_(self.plane, key="code")
        names = [f"{plcode}{ik:04}" for ik in self.ring.get_uint32_index(refpts)]
        super().exclude_vars(*varid, *names)

    @property
    def bpmweight(self) -> FloatArray:
        """Weight of position readings."""
        return self.observables[0].weight

    @bpmweight.setter
    def bpmweight(self, value: npt.ArrayLike):
        self.observables[0].weight = value

    @property
    def steerdelta(self) -> np.ndarray:
        """Step and weight on steerers."""
        return self.variables.deltas

    @steerdelta.setter
    def steerdelta(self, value):
        self.variables.deltas = value



"""Element translations and rotations.

.. Caution::

    The geometrical transformations are not commutative. When combining several
    transformations on the same element by using multiple function calls, the
    transformations are applied in a fixed order, independent of the order of the
    function calls:

    .. centered:: translations -> tilt (z-axis) -> yaw (y-axis) -> pitch (x-axis)
"""

from __future__ import annotations

__all__ = [
    "ReferencePoint",
    "get_offsets_rotations",
    "transform_elem",
    "shift_elem",
    "tilt_elem",
    "set_shift",
    "set_tilt",
    "set_rotation",
]

from enum import Enum
from collections.abc import Sequence

import numpy as np

from .elements import Element
from .utils import Refpts, All, refpts_iterator, _refcount

_x_axis = np.array([1.0, 0.0, 0.0])
_y_axis = np.array([0.0, 1.0, 0.0])
_z_axis = np.array([0.0, 0.0, 1.0])


# noinspection PyPep8Naming
def _rotation(rotations):
    """
    The implementation follows the one described in:
    https://doi.org/10.1016/j.nima.2022.167487
    All the comments featuring 'Eq' points to the paper's equations.

    3D rotation matrix (extrinsic rotation) using the Tait–Bryan angles
    convention.
    For more details, refer to https://en.wikipedia.org/wiki/Euler_angles
    Corresponds to Eq. (3)
    alpha: Rotation about the X-axis (pitch).
    beta: Rotation about the Y-axis (yaw).
    gamma: Rotation about the Z-axis (roll/tilt).
    """
    alpha, beta, gamma = rotations  # ZYX intrinsic rotations (pitch, yaw, tilt)
    R_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(alpha), -np.sin(alpha)],
            [0, np.sin(alpha), np.cos(alpha)],
        ]
    )
    R_y = np.array(
        [[np.cos(beta), 0, np.sin(beta)], [0, 1, 0], [-np.sin(beta), 0, np.cos(beta)]]
    )
    R_z = np.array(
        [
            [np.cos(gamma), -np.sin(gamma), 0],
            [np.sin(gamma), np.cos(gamma), 0],
            [0, 0, 1],
        ]
    )
    return R_x @ R_y @ R_z


# noinspection PyPep8Naming
def _translation_vector(ld, r3d, offsets, X_axis, Y_axis):
    """
    The implementation follows the one described in:
    https://doi.org/10.1016/j.nima.2022.167487
    All the comments featuring 'Eq' points to the paper's equations.

    Translation vector resulting from the joint effect of a longitudinal
    displacement (in the rotated frame), 3D offsets, and the 3D rotation matrix.
    Corresponds to Eqs. (8-11)
    ld: Longitudinal displacement [m]
    r3d: 3D rotation matrix
    offsets: 3D offsets [m]
    X_axis: X unit axis in rotated frame expressed in the xyz coordinate system
    Y_axis: Y unit axis in rotated frame expressed in the xyz coordinate system
    """
    tD0 = np.array([-offsets @ X_axis, 0, -offsets @ Y_axis, 0, 0, 0])
    T0 = np.array(
        [
            ld * r3d[2, 0] / r3d[2, 2],
            r3d[2, 0],
            ld * r3d[2, 1] / r3d[2, 2],
            r3d[2, 1],
            0,
            ld / r3d[2, 2],
        ]
    )
    return T0 + tD0


# noinspection PyPep8Naming
def _offsets_from_translation_vector(T, ld, r3d, X_axis, Y_axis, Z_axis):
    """
    Retrieve the 3D offsets from the T1 translation vector.
    T: Translation vector (T1)
    ld: Longitudinal displacement [m]
    r3d: 3D rotation matrix
    """
    T0 = np.array(
        [
            ld * r3d[2, 0] / r3d[2, 2],
            r3d[2, 0],
            ld * r3d[2, 1] / r3d[2, 2],
            r3d[2, 1],
            0,
            ld / r3d[2, 2],
        ]
    )
    tD0 = T - T0
    offsets = -tD0[0] * X_axis - tD0[2] * Y_axis + (ld + tD0[5]) * Z_axis

    return offsets


def _r_matrix(ld, r3d):
    """
    The implementation follows the one described in:
    https://doi.org/10.1016/j.nima.2022.167487
    All the comments featuring 'Eq' points to the paper's equations.

    Rotation matrix operator (R1, R2).
    Can take into account the effect of a longitudinal displacement.
    ld: Longitudinal displacement [m]
    r3d: 3D rotation matrix
    Corresponds to Eq. (9)
    """
    return np.array(
        [
            [
                r3d[1, 1] / r3d[2, 2],
                ld * r3d[1, 1] / r3d[2, 2] ** 2,
                -r3d[0, 1] / r3d[2, 2],
                -ld * r3d[0, 1] / r3d[2, 2] ** 2,
                0,
                0,
            ],
            [0, r3d[0, 0], 0, r3d[1, 0], r3d[2, 0], 0],
            [
                -r3d[1, 0] / r3d[2, 2],
                -ld * r3d[1, 0] / r3d[2, 2] ** 2,
                r3d[0, 0] / r3d[2, 2],
                ld * r3d[0, 0] / r3d[2, 2] ** 2,
                0,
                0,
            ],
            [0, r3d[0, 1], 0, r3d[1, 1], r3d[2, 1], 0],
            [0, 0, 0, 0, 1, 0],
            [
                -r3d[0, 2] / r3d[2, 2],
                -ld * r3d[0, 2] / r3d[2, 2] ** 2,
                -r3d[1, 2] / r3d[2, 2],
                -ld * r3d[1, 2] / r3d[2, 2] ** 2,
                0,
                1,
            ],
        ]
    )


def _ld_and_r3d_from_r_matrix(r_matrix):
    """
    Extracts the longitudinal displacement (ld) and 3D rotation matrix (r3d)
    from the R1 rotation matrix operator.
    r_matrix: 6x6 rotation matrix operator (R1).
    """
    r3d = np.eye(3)
    r3d[0, 0] = r_matrix[1, 1]
    r3d[1, 0] = r_matrix[1, 3]
    r3d[2, 0] = r_matrix[1, 4]
    r3d[0, 1] = r_matrix[3, 1]
    r3d[1, 1] = r_matrix[3, 3]
    r3d[2, 1] = r_matrix[3, 4]
    r3d[0, 2] = -r_matrix[5, 0] * r_matrix[1, 1] / r_matrix[2, 2]
    r3d[1, 2] = -r_matrix[5, 2] * r_matrix[1, 1] / r_matrix[2, 2]
    r3d[2, 2] = r_matrix[1, 1] / r_matrix[2, 2]
    ld = r_matrix[0, 1] * r_matrix[3, 3] / r_matrix[0, 0]
    return ld, r3d


class ReferencePoint(Enum):
    """Definition of the reference point for the geometric transformations."""

    CENTRE = "CENTRE"  #: Origin at the centre of the element.
    ENTRANCE = "ENTRANCE"  #: Origin at the entrance of the element.


# noinspection PyPep8Naming
def get_offsets_rotations(
    elem: Element,
    reference: ReferencePoint = ReferencePoint.CENTRE,
    *,
    RB_half: np.ndarray = None,
):
    r"""Return the offsets and rotations of a given element.

    This function returns the offsets [dx, dy, dz] and angular rotations (tilt, yaw,
    pitch) of the element.

    Args:
        elem: The beamline element.
        reference:  Transformation reference, either
                    :py:obj:`ReferencePoint.CENTRE` or
                    :py:obj:`ReferencePoint.ENTRANCE`. This must be identical to
                    the value used when the element was transformed.

    Returns:
        offsets (np.ndarray): [dx, dy, dz] array.
        tilt (float): Tilt angle (rotation about the Z-axis) in radians.
        yaw (float): Yaw angle (rotation about the Y-axis) in radians.
        pitch (float): Pitch angle (rotation about the X-axis) in radians.

    Raises:
        ValueError: If the `reference` argument is neither `ReferencePoint.CENTRE` nor
            `ReferencePoint.ENTRANCE`.

    .. Attention::

        The *reference* argument must be identical to the one used when displacing the
        element. Otherwise, the result is unpredictable.
    """
    if RB_half is None:
        elem_bending_angle = getattr(elem, "BendingAngle", 0.0)
        RB_half = _rotation([0.0, -elem_bending_angle / 2.0, 0.0])  # Eq. (12)

    try:
        T1 = elem.T1
    except AttributeError:
        T1 = np.zeros(6)
    try:
        R1 = elem.R1
    except AttributeError:
        R1 = np.eye(6)

    ld, r3d_tmp = _ld_and_r3d_from_r_matrix(R1)
    if reference is ReferencePoint.CENTRE:
        r3d = RB_half.T @ r3d_tmp @ RB_half

        X_axis = RB_half.T @ _x_axis
        Y_axis = RB_half.T @ _y_axis
        Z_axis = RB_half.T @ _z_axis

    elif reference is ReferencePoint.ENTRANCE:
        r3d = r3d_tmp

        X_axis = r3d @ _x_axis
        Y_axis = r3d @ _y_axis
        Z_axis = r3d @ _z_axis

    else:
        raise ValueError(
            "Unsupported reference, please choose either "
            "ReferencePoint.CENTRE or "
            "ReferencePoint.ENTRANCE."
        )

    offsets = _offsets_from_translation_vector(T1, ld, r3d, X_axis, Y_axis, Z_axis)
    tilt = np.arctan2(-r3d[0, 1], r3d[0, 0])
    yaw = np.arcsin(r3d[0, 2])
    pitch = np.arctan2(-r3d[1, 2], r3d[2, 2])
    return offsets, tilt, yaw, pitch


# noinspection PyPep8Naming
def transform_elem(
    elem: Element,
    reference: ReferencePoint = ReferencePoint.CENTRE,
    dx: float | None = None,
    dy: float | None = None,
    dz: float | None = None,
    tilt: float | None = None,
    pitch: float | None = None,
    yaw: float | None = None,
    *,
    relative: bool = False,
) -> None:
    r"""Set the translations and rotations of an :py:class:`.Element`.

    The tilt is a rotation around the *s*-axis, the pitch is a
    rotation around the *x*-axis, and the yaw is a rotation around the *y*-axis.

    A positive angle represents a clockwise rotation when
    looking in the direction of the rotation axis.

    The transformations are not all commutative. The translations are applied before
    the rotations. The rotations are applied in the order *Z* -> *Y* -> *X*
    (tilt -> yaw -> pitch). The element is rotated around its mid-point. The mid-point
    can either be the element entrance or its centre (axis joining the entry and exit
    points of the element).

    If *relative* is :py:obj:`True`, the previous translations and angles are rebuilt
    from the *r3d* matrix and incremented by the input arguments.

    PyAT describes the ultra-relativistic beam dynamics in 6D phase space
    coordinates, which differ from 3D spatial angles in an expansion with
    respect to the energy to first order by a factor :math:`(1 + \delta)`, where
    :math:`\delta` is the relative momentum deviation. This introduces
    spurious dispersion (angle proportional to :math:`\delta`).

    The implementation follows the one described in:
    https://doi.org/10.1016/j.nima.2022.167487
    All the comments featuring 'Eq' point to the paper's equations.

    Parameters:
        elem:           Element to be transformed.
        reference:      Transformation reference, either
                        :py:obj:`ReferencePoint.CENTRE` or
                        :py:obj:`ReferencePoint.ENTRANCE`.
        dx:             Horizontal shift [m]. Default: no change.
        dy:             Vertical shift [m]. Default: no change.
        dz:             Longitudinal shift [m]. Default: no change.
        tilt:           Tilt angle [rad]. Default: no change.
        pitch:          Pitch angle [rad]. Default: no change.
        yaw:            Yaw angle [rad]. Default: no change
        relative:       If :py:obj:`True`, the input values are added to the
          previous ones.

    .. Attention::

        When combining several transformations by using multiple calls to
        :py:func:`transform_elem`, the *reference* argument must be identical for all.
        Otherwise, the result is unpredictable.

    See Also:
        :py:func:`get_offsets_rotations`
    """
    if relative:

        def _set(ini, val):
            return ini if val is None else ini + val

    else:

        def _set(ini, val):
            return ini if val is None else val

    elem_length = getattr(elem, "Length", 0)
    elem_bending_angle = getattr(elem, "BendingAngle", 0)

    # Rotated transfer matrix linked to a bending angle
    RB = _rotation([0, -elem_bending_angle, 0])  # Eq. (12)
    RB_half = _rotation([0, -elem_bending_angle / 2, 0])  # Eq. (12)

    # Get the current transformation
    offsets0, tilt0, yaw0, pitch0 = get_offsets_rotations(
        elem, reference, RB_half=RB_half
    )

    # Apply new offsets and rotations (XYZ intrinsic order)
    offsets = np.array([_set(v0, v) for v0, v in zip(offsets0, [dx, dy, dz])])
    rotations = [_set(pitch0, pitch), _set(yaw0, yaw), _set(tilt0, tilt)]

    if reference is ReferencePoint.CENTRE:
        # Compute entrance rotation matrix in the rotated frame
        r3d_entrance = RB_half @ _rotation(rotations) @ RB_half.T  # Eq. (31)

        if elem_bending_angle:
            Rc = elem_length / elem_bending_angle
            OO0 = Rc * np.sin(elem_bending_angle / 2) * RB_half @ _z_axis  # Eq. (34)
            P0P = (
                -Rc * np.sin(elem_bending_angle / 2) * r3d_entrance @ RB_half @ _z_axis
            )  # Eq. (36)
        else:
            OO0 = elem_length / 2 * _z_axis  # Eq. (34)
            P0P = -elem_length / 2 * r3d_entrance @ _z_axis  # Eq. (36)

        # Transform offset to magnet entrance
        OP = OO0 + P0P + RB_half @ offsets  # Eq. (33)

    elif reference is ReferencePoint.ENTRANCE:
        r3d_entrance = _rotation(rotations)  # Eq. (3)
        OP = offsets  # Eq. (2)
    else:
        raise ValueError(
            "Unsupported reference, please choose either "
            "ReferencePoint.CENTRE or "
            "ReferencePoint.ENTRANCE."
        )

    # R1, T1
    # XYZ - axes unit - vectors expressed in the xyz coordinate system
    X_axis = r3d_entrance @ _x_axis
    Y_axis = r3d_entrance @ _y_axis
    Z_axis = r3d_entrance @ _z_axis

    ld_entrance = Z_axis @ OP  # Eq. (33)

    R1 = _r_matrix(ld_entrance, r3d_entrance)
    T1 = np.linalg.inv(R1) @ _translation_vector(
        ld_entrance, r3d_entrance, OP, X_axis, Y_axis
    )

    # R2, T2
    # XYZ - axes unit - vectors expressed in the xyz coordinate system
    X_axis = RB @ _x_axis
    Y_axis = RB @ _y_axis
    Z_axis = RB @ _z_axis

    r3d_exit = RB.T @ r3d_entrance.T @ RB  # Eq. (18) or (32)

    if elem_bending_angle:
        Rc = elem_length / elem_bending_angle
        OPp = np.array(
            [
                Rc * (np.cos(elem_bending_angle) - 1),
                0,
                elem_length * np.sin(elem_bending_angle) / elem_bending_angle,
            ]
        )  # Eq. (24)
    else:
        OPp = elem_length * _z_axis  # Eq. (24)

    OOp = r3d_entrance @ OPp + OP  # Eq. (25)
    OpPp = OPp - OOp

    ld_exit = Z_axis @ OpPp  # Eq. (23) or (37)

    R2 = _r_matrix(ld_exit, r3d_exit)
    T2 = _translation_vector(ld_exit, r3d_exit, OpPp, X_axis, Y_axis)

    # Update element
    elem.R1 = R1
    elem.R2 = R2
    elem.T1 = T1
    elem.T2 = T2


def tilt_elem(elem: Element, rots: float, relative: bool = False) -> None:
    r"""Set the tilt angle :math:`\theta` of an :py:class:`.Element`

    The rotation matrices are stored in the :pycode:`R1` and :pycode:`R2`
    attributes.

    :math:`R_1=\begin{pmatrix} cos\theta & sin\theta \\
    -sin\theta & cos\theta \end{pmatrix}`,
    :math:`R_2=\begin{pmatrix} cos\theta & -sin\theta \\
    sin\theta & cos\theta \end{pmatrix}`

    Parameters:
        elem:           Element to be tilted
        rots:           Tilt angle :math:`\theta` [rd]. *rots* > 0 corresponds
          to a corkscrew rotation of the element looking in the direction of
          the beam. Use :py:obj:`None` to keep the current value.
        relative:       If :py:obj:`True`, the rotation is added to the
          existing one

    See Also:
        :py:func:`shift_elem`
        :py:func:`.transform_elem`
    """
    transform_elem(elem, tilt=rots, relative=relative)


def shift_elem(
    elem: Element,
    dx: float | None = 0.0,
    dy: float | None = 0.0,
    dz: float | None = 0.0,
    *,
    relative: bool = False,
) -> None:
    r"""Sets the translations of an :py:class:`.Element`

    The translation vectors are stored in the :pycode:`T1` and :pycode:`T2`
    attributes.

    Parameters:
        elem:           Element to be shifted
        dx:             Horizontal translation [m]. Use :py:obj:`None` to keep
          the current value.
        dy:             Vertical translation [m]. Use :py:obj:`None` to keep
          the current value.
        dz:             Longitudinal translation [m]. Use :py:obj:`None` to keep
          the current value.
        relative:       If :py:obj:`True`, the translation is added to the
          existing one

    See Also:
        :py:func:`tilt_elem`
        :py:func:`.transform_elem`
    """
    transform_elem(elem, dx=dx, dy=dy, dz=dz, relative=relative)


def set_rotation(
    ring: Sequence[Element],
    tilts=0.0,
    pitches=0.0,
    yaws=0.0,
    *,
    refpts: Refpts = All,
    relative=False,
) -> None:
    r"""Sets the rotations of a list of elements.

    Parameters:
        ring:       Lattice description.
        tilts:      Scalar or Sequence of tilt values applied to the
          selected elements. Use :py:obj:`None` to keep the current values.
        pitches:    Scalar or Sequence of pitch values applied to the
          selected elements. Use :py:obj:`None` to keep the current values.
        yaws:       Scalar or Sequence of yaw values applied to the
          selected elements. Use :py:obj:`None` to keep the current values.
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        relative:   If :py:obj:`True`, the rotations are added to the existing ones.

    .. versionadded:: 0.7.0
       The *refpts* argument

    See Also:
        :py:func:`set_tilt`
        :py:func:`set_shift`
    """
    nb = _refcount(ring, refpts, endpoint=False)
    tilts = np.broadcast_to(tilts, (nb,))
    pchs = np.broadcast_to(pitches, (nb,))
    yaws = np.broadcast_to(yaws, (nb,))
    for el, tilt, pitch, yaw in zip(refpts_iterator(ring, refpts), tilts, pchs, yaws):
        transform_elem(el, tilt=tilt, pitch=pitch, yaw=yaw, relative=relative)


def set_tilt(
    ring: Sequence[Element], tilts, *, refpts: Refpts = All, relative=False
) -> None:
    r"""Sets the tilts of a list of elements.

    Parameters:
        ring:       Lattice description.
        tilts:      Scalar or Sequence of tilt values applied to the
          selected elements. Use :py:obj:`None` to keep the current values.
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        relative:   If :py:obj:`True`, the rotation is added to the existing one.

    .. versionadded:: 0.7.0
       The *refpts* argument

    See Also:
        :py:func:`set_rotation`
        :py:func:`set_shift`
    """
    nb = _refcount(ring, refpts, endpoint=False)
    tilts = np.broadcast_to(tilts, (nb,))
    for el, tilt in zip(refpts_iterator(ring, refpts), tilts):
        transform_elem(el, tilt=tilt, relative=relative)


def set_shift(
    ring: Sequence[Element], dxs, dys, dzs=None, *, refpts: Refpts = All, relative=False
) -> None:
    r"""Sets the translations of a list of elements.

    Parameters:
        ring:       Lattice description.
        dxs:        Scalar or Sequence of horizontal translations values applied
          to the selected elements. Use :py:obj:`None` to keep the current values [m].
        dys:        Scalar or Sequence of vertical translations values applied
          to the selected elements. Use :py:obj:`None` to keep the current values [m].
        dzs:        Scalar or Sequence of longitudinal translations values applied
          to the selected elements. Use :py:obj:`None` to keep the current values [m].
        refpts:     Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        relative:   If :py:obj:`True`, the translation is added to the
          existing one.

    .. versionadded:: 0.7.0
       The *refpts* argument

    See Also:
        :py:func:`set_rotation`
        :py:func:`set_tilt`
    """
    nb = _refcount(ring, refpts, endpoint=False)
    dxs = np.broadcast_to(dxs, (nb,))
    dys = np.broadcast_to(dys, (nb,))
    dzs = np.broadcast_to(dzs, (nb,))
    for el, dx, dy, dz in zip(refpts_iterator(ring, refpts), dxs, dys, dzs):
        transform_elem(el, dx=dx, dy=dy, dz=dz, relative=relative)


def _get_dx(elem: Element) -> float:
    """Horizontal element shift"""
    offsets, _, _, _ = get_offsets_rotations(elem, ReferencePoint.CENTRE)
    return offsets[0]


def _set_dx(elem: Element, value: float) -> None:
    transform_elem(elem, ReferencePoint.CENTRE, dx=value)


def _get_dy(elem: Element) -> float:
    """Vertical element shift"""
    offsets, _, _, _ = get_offsets_rotations(elem, ReferencePoint.CENTRE)
    return offsets[1]


def _set_dy(elem: Element, value: float) -> None:
    transform_elem(elem, ReferencePoint.CENTRE, dy=value)


def _get_dz(elem: Element) -> float:
    """Longitudinal element shift"""
    offsets, _, _, _ = get_offsets_rotations(elem, ReferencePoint.CENTRE)
    return offsets[2]


def _set_dz(elem: Element, value: float) -> None:
    transform_elem(elem, ReferencePoint.CENTRE, dz=value)


def _get_tilt(elem: Element) -> float:
    """Element tilt"""
    _, tilt, _, _ = get_offsets_rotations(elem, ReferencePoint.CENTRE)
    return tilt


def _set_tilt(elem: Element, value: float) -> None:
    transform_elem(elem, ReferencePoint.CENTRE, tilt=value)


Element.transform = transform_elem
Element.dx = property(_get_dx, _set_dx)
Element.dy = property(_get_dy, _set_dy)
Element.dz = property(_get_dz, _set_dz)
Element.tilt = property(_get_tilt, _set_tilt)



from __future__ import annotations

__all__ = ["get_energy_loss", "set_cavity_phase", "ELossMethod", "get_timelag_fromU0"]

from enum import Enum
from warnings import warn
from collections.abc import Sequence

import numpy as np
from scipy.optimize import least_squares

from at.constants import clight, Cgamma
from at.lattice import Lattice, Dipole, Wiggler, RFCavity, Refpts, EnergyLoss
from at.lattice import check_radiation, AtError, AtWarning
from at.lattice import get_bool_index, set_value_refpts
from at.lattice import DConstant


class ELossMethod(Enum):
    """methods for the computation of energy losses"""

    #: The losses are obtained from
    #: :math:`E_{loss}=C_\gamma/2\pi . E^4 . I_2`.
    #: Takes into account bending magnets and wigglers.
    INTEGRAL = 1
    #: The losses are obtained by tracking without cavities.
    #: Needs radiation ON, takes into account all radiating elements
    TRACKING = 2


def get_energy_loss(
    ring: Lattice, method: ELossMethod | None = ELossMethod.INTEGRAL
) -> float:
    """Computes the energy loss per turn

    Parameters:
        ring:           Lattice description
        method:         Method for energy loss computation.
          See :py:class:`ELossMethod`.

    Returns:
        eloss (float):  Energy loss per turn [eV]
    """

    # noinspection PyShadowingNames
    def integral(ring):
        """Losses = Cgamma / 2pi * EGeV^4 * i2"""

        def wiggler_i2(wiggler: Wiggler):
            rhoinv = wiggler.Bmax / ring.BRho
            coefh = wiggler.By[1, :]
            coefv = wiggler.Bx[1, :]
            return (
                wiggler.Length
                * (np.sum(coefh * coefh) + np.sum(coefv * coefv))
                * rhoinv**2
                / 2
            )

        def dipole_i2(dipole: Dipole):
            return dipole.BendingAngle**2 / dipole.Length

        def eloss_i2(eloss: EnergyLoss):
            return eloss.EnergyLoss / coef

        i2 = 0.0
        coef = Cgamma / 2.0 / np.pi * ring.energy**4
        for el in ring:
            if isinstance(el, Dipole):
                i2 += dipole_i2(el)
            elif isinstance(el, Wiggler) and el.PassMethod != "DriftPass":
                i2 += wiggler_i2(el)
            elif isinstance(el, EnergyLoss) and el.PassMethod != "IdentityPass":
                i2 += eloss_i2(el)
        e_loss = coef * i2
        return e_loss

    # noinspection PyShadowingNames
    @check_radiation(True)
    def tracking(ring):
        """Losses from tracking"""
        energy = ring.energy
        particle = ring.particle
        delta = 0.0
        for e in ring:
            if e.PassMethod.endswith("RadPass"):
                ot = e.track(np.zeros(6), energy=energy, particle=particle)
                delta += ot[4]
        return -delta * energy

    if isinstance(method, str):
        method = ELossMethod[method.upper()]
        warn(FutureWarning(f"You should use {method!s}"), stacklevel=2)
    if method is ELossMethod.INTEGRAL:
        return ring.periodicity * integral(ring)
    elif method == ELossMethod.TRACKING:
        return ring.periodicity * tracking(ring)
    else:
        raise AtError(f"Invalid method: {method}")


# noinspection PyPep8Naming
def get_timelag_fromU0(
    ring: Lattice,
    *,
    method: ELossMethod | None = ELossMethod.TRACKING,
    cavpts: Refpts | None = None,
    divider: int | None = 4,
    ts_tol: float | None = None,
) -> tuple[Sequence[float], float]:
    """
    Get the TimeLag attribute of RF cavities based on frequency,
    voltage and energy loss per turn, so that the synchronous phase is zero.
    Used in set_cavity_phase()

    Parameters:
        ring:               Lattice description
        method:             Method for energy loss computation.
          See :py:class:`ELossMethod`.
        cavpts:             Cavity location. If None, use all cavities.
          This allows to ignore harmonic cavities.
        divider: number of segments to search for ts
        ts_tol: relative tolerance for ts calculation, default is 1.0-9
          If the search fails to find the synchronous phase the tolerance
          can be increased using the keyword argument or by setting
          `at.DConstant.TStol`
    Returns:
        timelag (float):    (ncav,) array of *Timelag* values
        ts (float):         Time difference with the present value
    """

    def singlev(values):
        vals = np.unique(values)
        if len(vals) > 1:
            raise AtError("values not equal for all cavities")
        return vals[0]

    def eq(x, freq, rfv, tl0, u0):
        omf = 2 * np.pi * freq / clight
        if u0 > 0.0:
            eq1 = (np.sum(-rfv * np.sin(omf * (x - tl0))) - u0) / u0
        else:
            eq1 = np.sum(-rfv * np.sin(omf * (x - tl0)))
        eq2 = np.sum(-omf * rfv * np.cos(omf * (x - tl0)))
        if eq2 > 0:
            return np.sqrt(eq1**2 + eq2**2)
        else:
            return abs(eq1)

    if cavpts is None:
        cavpts = get_bool_index(ring, RFCavity)
    if ts_tol is None:
        ts_tol = DConstant.TStol
    u0 = get_energy_loss(ring, method=method) / ring.periodicity
    freq = np.array([cav.Frequency for cav in ring.select(cavpts)])
    rfv = np.array([cav.Voltage for cav in ring.select(cavpts)])
    tl0 = np.array([cav.TimeLag for cav in ring.select(cavpts)])
    try:
        frf = singlev(freq)
        tml = singlev(tl0)
    except AtError:
        ctmax = clight / np.amin(freq) / 2
        tt0 = tl0[np.argmin(freq)]
        bounds = (-ctmax, ctmax)
        args = (freq, rfv, tl0, u0)
        r = []
        for i in range(divider):
            fact = (i + 1) / divider
            r.append(
                least_squares(
                    eq, bounds[0] * fact + tt0, args=args, bounds=bounds + tt0
                )
            )
            r.append(
                least_squares(
                    eq, bounds[1] * fact + tt0, args=args, bounds=bounds + tt0
                )
            )
        res = np.array([ri.fun[0] for ri in r])
        ok = res < ts_tol
        vals = np.array([abs(ri.x[0]).round(decimals=6) for ri in r])
        if not np.any(ok):
            raise AtError("No solution found for Phis: check RF settings") from None
        if len(np.unique(vals[ok])) > 1:
            warn(
                AtWarning("More than one solution found for Phis: check RF settings"),
                stacklevel=2,
            )
        ts = -r[np.argmin(res)].x[0]
        timelag = ts + tl0
    else:
        vrf = np.sum(rfv)
        if u0 > vrf:
            v1 = ring.periodicity * vrf
            v2 = ring.periodicity * u0
            raise AtError(
                f"The RF voltage ({v1:.3e} eV) is lower than "
                f"the radiation losses ({v2:.3e} eV)."
            )
        timelag = clight / (2 * np.pi * frf) * np.arcsin(u0 / vrf)
        ts = timelag - tml
        timelag *= np.ones(ring.refcount(cavpts))
    return timelag, ts


def set_cavity_phase(
    ring: Lattice,
    *,
    method: ELossMethod = ELossMethod.TRACKING,
    refpts: Refpts | None = None,
    cavpts: Refpts | None = None,
    copy: bool = False,
) -> None:
    """
    Adjust the TimeLag attribute of RF cavities based on frequency,
    voltage and energy loss per turn, so that the synchronous phase is zero.
    An error occurs if all cavities do not have the same frequency.

    .. Warning::

       This function changes the time reference, this should be avoided

    Parameters:
        ring:       Lattice description
        method:     Method for energy loss computation.
          See :py:class:`ELossMethod`.
        cavpts:     Cavity location. If None, use all cavities.
          This allows to ignore harmonic cavities.
        copy:       If True, returns a shallow copy of ring with new
          cavity elements. Otherwise, modify ring in-place.
    """
    # refpts is kept for backward compatibility
    if cavpts is None and refpts is not None:
        warn(FutureWarning('You should use "cavpts" instead of "refpts"'), stacklevel=2)
        cavpts = refpts
    elif cavpts is None:
        cavpts = get_bool_index(ring, RFCavity)
    timelag, _ = get_timelag_fromU0(ring, method=method, cavpts=cavpts)
    set_value_refpts(ring, cavpts, "TimeLag", timelag, copy=copy)


Lattice.get_energy_loss = get_energy_loss
Lattice.energy_loss = property(get_energy_loss)
Lattice.set_cavity_phase = set_cavity_phase



import numpy as np
from numpy.testing import assert_allclose as assert_close

import at
from at import (
    RingObservable,
    ObservableList,
    OrbitObservable,
    GlobalOpticsObservable,
    LocalOpticsObservable,
    MatrixObservable,
    TrajectoryObservable,
    EmittanceObservable,
    LatticeObservable,
    GeometryObservable,
)


def test_observables(hmba_lattice):
    # noinspection PyUnusedLocal
    def phase_advance(elemdata):
        mu = elemdata.mu
        return mu[-1] - mu[0]

    def circumference(ring):
        return ring.cell_length

    ring = hmba_lattice.enable_6d(copy=True)
    ring.set_rf_frequency(dp=0.0)

    # Create an empty ObervableList
    allobs = ObservableList()
    # Populate with all kinds of Observables
    allobs.append(OrbitObservable(at.Monitor, axis="x"))
    allobs.append(
        LocalOpticsObservable(
            at.Monitor, "beta", plane=1, target=7.0, bounds=(-np.inf, 0.0)
        )
    )
    allobs.append(MatrixObservable("BPM_02"))
    allobs.append(LocalOpticsObservable(at.Monitor, "beta", plane="v", statfun=np.amax))
    allobs.append(
        LocalOpticsObservable(
            at.Quadrupole, "closed_orbit", plane=slice(4), target=0.0, weight=1.0e-6
        )
    )
    allobs.append(LocalOpticsObservable(at.Quadrupole, "s_pos"))
    allobs.append(
        LocalOpticsObservable([33, 101], phase_advance, all_points=True, summary=True)
    )
    allobs.append(GlobalOpticsObservable("tune", plane=0, use_integer=True))
    allobs.append(LocalOpticsObservable(at.End, "mu"))
    allobs.append(GlobalOpticsObservable("chromaticity"))
    allobs.append(LatticeObservable(at.Sextupole, "H", statfun=np.mean))
    allobs.append(LatticeObservable(at.Sextupole, "PolynomB", index=2))
    allobs.append(EmittanceObservable("emittances", plane="x"))
    allobs.append(RingObservable(circumference))
    allobs.append(TrajectoryObservable(at.Monitor, axis="px"))
    allobs.append(GeometryObservable(at.Monitor, "x"))

    # Evaluate the Observables
    r_in = np.zeros(6)
    r_in[0] = 0.001
    r_in[2] = 0.001
    allobs.evaluate(ring, r_in=r_in, initial=True)

    # Get the expected values
    o0, o = ring.find_orbit(refpts=at.All)
    el0, rg, el = ring.get_optics(refpts=at.All, orbit=o0, get_chrom=True)
    m66, ms = ring.find_m66(refpts="BPM_02", orbit=o0)
    prms = ring.envelope_parameters(orbit=o0)
    rout, _, _ = ring.track(r_in, refpts=at.Monitor)
    geodata, _ = ring.get_geometry(refpts=at.Monitor)

    monitors = ring.get_bool_index(at.Monitor)
    quadrupoles = ring.get_bool_index(at.Quadrupole)

    # Compare the results
    assert_close(allobs.values[0], o[monitors, 0])
    assert_close(allobs.values[1], el.beta[monitors, 1])
    assert_close(allobs.values[2], ms)
    assert_close(allobs.values[3], np.amax(el.beta[monitors, 1]))
    assert_close(allobs.values[4], el.closed_orbit[quadrupoles, :4])
    assert_close(allobs.values[5], el.s_pos[quadrupoles])
    assert_close(allobs.values[6], el.mu[101] - el.mu[33])
    assert_close(allobs.values[7], el.mu[-1, 0] / 2.0 / np.pi)
    assert_close(allobs.values[8], [el.mu[-1]])
    assert_close(allobs.values[9], rg.chromaticity)
    assert_close(
        allobs.values[10], np.mean([elem.H for elem in ring.select(at.Sextupole)])
    )
    assert_close(
        allobs.values[11], [elem.PolynomB[2] for elem in ring.select(at.Sextupole)]
    )
    assert_close(allobs.values[12], prms.emittances[0])
    assert_close(allobs.values[13], ring.cell_length)
    assert_close(allobs.values[14], rout[1, 0, :, 0])
    assert_close(allobs.values[15], geodata.x)



import sys
import os
from os.path import basename, exists, join, splitext
import glob

# Numpy build dependency defined in pyproject.toml.
import numpy as np
from setuptools import setup, Extension


def select_omp():
    if os.uname().machine.startswith("arm"):
        homeb = "/opt/homebrew/opt/libomp"
    else:
        homeb = "/usr/local"

    if exists(os.path.join(homeb, "include/omp.h")):  # Homebrew
        return "-I" + os.path.join(homeb, "include"), os.path.join(homeb, "lib")
    elif exists("/opt/local/include/libomp/omp.h"):  # MacPorts
        return "-I/opt/local/include/libomp", "/opt/local/lib/libomp"
    else:
        raise FileNotFoundError(
            "\n".join(
                (
                    "",
                    "libomp.dylib must be installed with your package manager:",
                    "",
                    'Use "$ brew install libomp"',
                    'Or  "$ sudo port install libomp"',
                    "",
                )
            )
        )


print("** Entering setup.py:", str(sys.argv))
print("** MPI:", os.environ.get("MPI", None))
print("** OPENMP:", os.environ.get("OPENMP", None))
print("** CUDA:", os.environ.get("CUDA", None))
print("** OPENCL:", os.environ.get("OPENCL", None))

platform = sys.platform
macros = [("PYAT", None)]

cflags = []
cppflags = []

if not platform.startswith("win32"):
    cflags += ["-std=c99", "-Wno-unused-function", "-Wno-unknown-pragmas"]
    cppflags += ["-Wno-unused-function", "-Wno-unknown-pragmas"]

if platform.startswith("darwin"):
    cflags += ["-ffp-contract=off"]
    cppflags += ["-ffp-contract=off"]

mpi = eval(os.environ.get("MPI", "None"))
if not mpi or (len(sys.argv) >= 2 and sys.argv[1] in {"egg_info", "sdist"}):
    mpi_macros = []
    mpi_includes = []
else:
    mpi_macros = [("MPI", None)]
    os.environ["CC"] = "mpicc"
    os.environ["CXX"] = "mpicxx"
    try:
        import mpi4py
    except ImportError:
        print(
            "\npyAT with MPI requires mpi4py. "
            'Please install mpi4py: "pip install mpi4py"\n'
        )
        sys.exit()
    mpi_includes = mpi4py.get_include()


omp = eval(os.environ.get("OPENMP", "None"))
if not omp:
    omp_cflags = []
    omp_lflags = []
    omp_macros = []
else:
    # Get the location of an alternate OpenMP library
    # Example: OMP_MATLAB=$MATLABROOT/sys/os/glnx64
    omp_path = os.environ.get("OMP_MATLAB", None)
    # Get the threshold on the number of particles
    omp_threshold = int(os.environ.get("OMP_PARTICLE_THRESHOLD", 10))
    omp_macros = [("OMP_PARTICLE_THRESHOLD", omp_threshold)]
    if platform.startswith("win"):
        omp_cflags = ["/openmp"]
        omp_lflags = []
    elif platform.startswith("darwin"):
        omp_inc, omp_lib = select_omp()
        omp_cflags = ["-Xpreprocessor", "-fopenmp", omp_inc]
        if omp_path is None:
            omp_lflags = ["-L" + omp_lib, "-lomp"]
        else:
            omp_lflags = ["-L" + omp_path, "-Wl,-rpath," + omp_path, "-liomp5"]
    else:
        omp_cflags = ["-fopenmp"]
        if omp_path is None:
            omp_lflags = ["-lgomp"]
        else:
            omp_lflags = ["-L" + omp_path, "-Wl,-rpath," + omp_path, "-liomp5"]


cuda = eval(os.environ.get("CUDA", "None"))
if not cuda:
    cuda_cppflags = []
    cuda_lflags = []
    cuda_macros = []
else:
    # Generate the shared include file for the GPU kernel
    exec(open("atgpu/genheader.py").read())
    cuda_path = os.environ.get("CUDA_PATH", None)
    cuda_macros = [("CUDA", None)]
    if cuda_path is None:
        raise RuntimeError("CUDA_PATH environment variable not defined")
    if sys.platform.startswith("win"):
        cuda_cppflags = ["-I" + cuda_path + "\\include"]
        cuda_lflags = ["/LIBPATH:" + cuda_path + "\\lib\\x64", "cuda.lib", "nvrtc.lib"]
    else:
        cuda_cppflags = ["-I" + cuda_path + "/include"]
        cuda_lflags = [
            "-L" + cuda_path + "/lib64",
            "-Wl,-rpath," + cuda_path + "/lib64",
            "-lcuda",
            "-lnvrtc",
        ]

opencl = eval(os.environ.get("OPENCL", "None"))
if not opencl:
    opencl_cppflags = []
    opencl_lflags = []
    opencl_macros = []
else:
    # Generate the shared include file for the GPU kernel
    exec(open("atgpu/genheader.py").read())
    opencl_ocl_path = os.environ.get("OCL_PATH", None)
    opencl_macros = [("OPENCL", None)]
    if sys.platform.startswith("win"):
        # Static link
        if opencl_ocl_path is None:
            raise RuntimeError("OCL_PATH environment variable not defined")
        opencl_cppflags = ["-I" + opencl_ocl_path + "\\include"]
        opencl_lflags = [
            "/LIBPATH:" + opencl_ocl_path + "\\lib",
            "OpenCL.lib",
            "cfgmgr32.lib",
            "runtimeobject.lib",
            "Advapi32.lib",
            "ole32.lib",
        ]
    elif sys.platform.startswith("darwin"):
        opencl_cppflags = ["-std=c++11"]
        opencl_lflags = ["-Wl,-framework,OpenCL"]
    else:
        if opencl_ocl_path is not None:
            # Private install
            opencl_cppflags = ["-I" + opencl_ocl_path + "/include"]
            opencl_lflags = [
                "-L" + opencl_ocl_path + "/lib",
                "-Wl,-rpath," + opencl_ocl_path + "/lib",
                "-lOpenCL",
            ]
        elif exists("/usr/include/CL/opencl.h"):
            # Standard install
            opencl_cppflags = []
            opencl_lflags = ["-lOpenCL"]
        else:
            raise RuntimeError(
                "Install OpenCL include and driver (ICD) in standard path or set OCL_PATH environment variable"
            )

if not sys.platform.startswith("win32"):
    cflags += ["-Wno-unused-function"]

# It is easier to copy the integrator files into a directory inside pyat
# for packaging. However, we cannot always rely on this directory being
# inside a clone of at and having the integrator sources available, and
# this file is executed each time any setup.py command is run.
# It appears that only copying the files when they are available is
# sufficient.
integrator_src_orig = "atintegrators"
diffmatrix_orig = join("atmat", "atphysics", "Radiation")

c_pass_methods = glob.glob(join(integrator_src_orig, "*Pass.c"))
cpp_pass_methods = glob.glob(join(integrator_src_orig, "*Pass.cc"))
gpu_pass_methods = glob.glob(join("atgpu", "*Pass.cpp"))
diffmatrix_source = join(diffmatrix_orig, "findmpoleraddiffmatrix.c")
at_source = join("pyat", "at.c")


def integrator_ext(pass_method, flags):
    name, _ = splitext(basename(pass_method))
    name = ".".join(("at", "integrators", name))
    return Extension(
        name=name,
        sources=[pass_method],
        include_dirs=[np.get_include(), mpi_includes, integrator_src_orig],
        define_macros=macros + omp_macros + mpi_macros,
        extra_compile_args=flags + omp_cflags,
        extra_link_args=omp_lflags,
    )


at = Extension(
    "at.tracking.atpass",
    sources=[at_source],
    define_macros=macros + omp_macros + mpi_macros,
    include_dirs=[np.get_include(), integrator_src_orig],
    extra_compile_args=cflags + omp_cflags,
    extra_link_args=omp_lflags,
)

cconfig = Extension(
    "at.cconfig",
    sources=[join("pyat", "at", "cconfig.c")],
    define_macros=macros + omp_macros + mpi_macros + cuda_macros + opencl_macros,
    extra_compile_args=cflags + omp_cflags,
)

diffmatrix = Extension(
    name="at.physics.diffmatrix",
    sources=[join(diffmatrix_orig, "findmpoleraddiffmatrix.c")],
    include_dirs=[np.get_include(), integrator_src_orig],
    define_macros=macros,
    extra_compile_args=cflags,
)

wiggdiffmatrix = Extension(
    name="at.physics.wiggdiffmatrix",
    sources=[join(diffmatrix_orig, "FDW.c")],
    include_dirs=[np.get_include(), integrator_src_orig],
    define_macros=macros,
    extra_compile_args=cflags,
)

gpusource = gpu_pass_methods + [
    join("atgpu", "AbstractGPU.cpp"),
    join("atgpu", "AbstractInterface.cpp"),
    join("atgpu", "PyInterface.cpp"),
    join("atgpu", "Lattice.cpp"),
    join("atgpu", "PassMethodFactory.cpp"),
    join("atgpu", "SymplecticIntegrator.cpp"),
]

cudaext = Extension(
    name="at.tracking.gpu",
    sources=gpusource + [join("atgpu", "CudaGPU.cpp")],
    define_macros=macros + cuda_macros,
    include_dirs=[np.get_include()],
    extra_compile_args=cppflags + cuda_cppflags,
    extra_link_args=cuda_lflags,
)

openclext = Extension(
    name="at.tracking.gpu",
    sources=gpusource + [join("atgpu", "OpenCLGPU.cpp")],
    define_macros=macros + opencl_macros,
    include_dirs=[np.get_include()],
    extra_compile_args=cppflags + opencl_cppflags,
    extra_link_args=opencl_lflags,
)

setup(
    ext_modules=[at, cconfig, diffmatrix, wiggdiffmatrix]
    + [integrator_ext(pm, cflags) for pm in c_pass_methods]
    + [integrator_ext(pm, cppflags) for pm in cpp_pass_methods]
    + ([cudaext] if cuda else [])
    + ([openclext] if opencl else []),
)

print("** Leaving setup.py")
